"""Session implementation for CherryPy.

You need to edit your config file to use sessions. Here's an example::

    [/]
    tools.sessions.on = True
    tools.sessions.storage_class = cherrypy.lib.sessions.FileSession
    tools.sessions.storage_path = "/home/site/sessions"
    tools.sessions.timeout = 60

This sets the session to be stored in files in the directory
/home/site/sessions, and the session timeout to 60 minutes. If you omit
``storage_class``, the sessions will be saved in RAM.
``tools.sessions.on`` is the only required line for working sessions,
the rest are optional.

By default, the session ID is passed in a cookie, so the client's browser must
have cookies enabled for your site.

To set data for the current session, use
``cherrypy.session['fieldname'] = 'fieldvalue'``;
to get data use ``cherrypy.session.get('fieldname')``.

================
Locking sessions
================

By default, the ``'locking'`` mode of sessions is ``'implicit'``, which means
the session is locked early and unlocked late. Be mindful of this default mode
for any requests that take a long time to process (streaming responses,
expensive calculations, database lookups, API calls, etc), as other concurrent
requests that also utilize sessions will hang until the session is unlocked.

If you want to control when the session data is locked and unlocked,
set ``tools.sessions.locking = 'explicit'``. Then call
``cherrypy.session.acquire_lock()`` and ``cherrypy.session.release_lock()``.
Regardless of which mode you use, the session is guaranteed to be unlocked when
the request is complete.

=================
Expiring Sessions
=================

You can force a session to expire with :func:`cherrypy.lib.sessions.expire`.
Simply call that function at the point you want the session to expire, and it
will cause the session cookie to expire client-side.

===========================
Session Fixation Protection
===========================

If CherryPy receives, via a request cookie, a session id that it does not
recognize, it will reject that id and create a new one to return in the
response cookie. This `helps prevent session fixation attacks
<http://en.wikipedia.org/wiki/Session_fixation#Regenerate_SID_on_each_request>`_.
However, CherryPy "recognizes" a session id by looking up the saved session
data for that id. Therefore, if you never save any session data,
**you will get a new session id for every request**.

A side effect of CherryPy overwriting unrecognised session ids is that if you
have multiple, separate CherryPy applications running on a single domain (e.g.
on different ports), each app will overwrite the other's session id because by
default they use the same cookie name (``"session_id"``) but do not recognise
each others sessions. It is therefore a good idea to use a different name for
each, for example::

    [/]
    ...
    tools.sessions.name = "my_app_session_id"

================
Sharing Sessions
================

If you run multiple instances of CherryPy (for example via mod_python behind
Apache prefork), you most likely cannot use the RAM session backend, since each
instance of CherryPy will have its own memory space. Use a different backend
instead, and verify that all instances are pointing at the same file or db
location. Alternately, you might try a load balancer which makes sessions
"sticky". Google is your friend, there.

================
Expiration Dates
================

The response cookie will possess an expiration date to inform the client at
which point to stop sending the cookie back in requests. If the server time
and client time differ, expect sessions to be unreliable. **Make sure the
system time of your server is accurate**.

CherryPy defaults to a 60-minute session timeout, which also applies to the
cookie which is sent to the client. Unfortunately, some versions of Safari
("4 public beta" on Windows XP at least) appear to have a bug in their parsing
of the GMT expiration date--they appear to interpret the date as one hour in
the past. Sixty minutes minus one hour is pretty close to zero, so you may
experience this bug as a new session id for every request, unless the requests
are less than one second apart. To fix, try increasing the session.timeout.

On the other extreme, some users report Firefox sending cookies after their
expiration date, although this was on a system with an inaccurate system time.
Maybe FF doesn't trust system time.
"""
import sys
import datetime
import os
import time
import threading
import binascii
import pickle

import zc.lockfile

import cherrypy
from cherrypy.lib import httputil
from cherrypy.lib import locking
from cherrypy.lib import is_iterator


missing = object()


class Session(object):
    """A CherryPy dict-like Session object (one per request)."""

    _id = None

    id_observers = None
    "A list of callbacks to which to pass new id's."

    @property
    def id(self):
        """Return the current session id."""
        return self._id

    @id.setter
    def id(self, value):
        self._id = value
        for o in self.id_observers:
            o(value)

    timeout = 60
    'Number of minutes after which to delete session data.'

    locked = False
    """
    If True, this session instance has exclusive read/write access
    to session data."""

    loaded = False
    """If True, data has been retrieved from storage.

    This should happen automatically on the first attempt to access
    session data.
    """

    clean_thread = None
    'Class-level Monitor which calls self.clean_up.'

    clean_freq = 5
    'The poll rate for expired session cleanup in minutes.'

    originalid = None
    'The session id passed by the client. May be missing or unsafe.'

    missing = False
    'True if the session requested by the client did not exist.'

    regenerated = False
    """True if the application called session.regenerate().

    This is not set by internal calls to regenerate the session id.
    """

    debug = False
    'If True, log debug information.'

    # --------------------- Session management methods --------------------- #

    def __init__(self, id=None, **kwargs):
        self.id_observers = []
        self._data = {}

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.originalid = id
        self.missing = False
        if id is None:
            if self.debug:
                cherrypy.log('No id given; making a new one', 'TOOLS.SESSIONS')
            self._regenerate()
        else:
            self.id = id
            if self._exists():
                if self.debug:
                    cherrypy.log('Set id to %s.' % id, 'TOOLS.SESSIONS')
            else:
                if self.debug:
                    cherrypy.log('Expired or malicious session %r; '
                                 'making a new one' % id, 'TOOLS.SESSIONS')
                # Expired or malicious session. Make a new one.
                # See https://github.com/cherrypy/cherrypy/issues/709.
                self.id = None
                self.missing = True
                self._regenerate()

    def now(self):
        """Generate the session specific concept of 'now'.

        Other session providers can override this to use alternative,
        possibly timezone aware, versions of 'now'.
        """
        return datetime.datetime.now()

    def regenerate(self):
        """Replace the current session (with a new id)."""
        self.regenerated = True
        self._regenerate()

    def _regenerate(self):
        if self.id is not None:
            if self.debug:
                cherrypy.log(
                    'Deleting the existing session %r before '
                    'regeneration.' % self.id,
                    'TOOLS.SESSIONS')
            self.delete()

        old_session_was_locked = self.locked
        if old_session_was_locked:
            self.release_lock()
            if self.debug:
                cherrypy.log('Old lock released.', 'TOOLS.SESSIONS')

        self.id = None
        while self.id is None:
            self.id = self.generate_id()
            # Assert that the generated id is not already stored.
            if self._exists():
                self.id = None
        if self.debug:
            cherrypy.log('Set id to generated %s.' % self.id,
                         'TOOLS.SESSIONS')

        if old_session_was_locked:
            self.acquire_lock()
            if self.debug:
                cherrypy.log('Regenerated lock acquired.', 'TOOLS.SESSIONS')

    def clean_up(self):
        """Clean up expired sessions."""
        pass

    def generate_id(self):
        """Return a new session id."""
        return binascii.hexlify(os.urandom(20)).decode('ascii')

    def save(self):
        """Save session data."""
        try:
            # If session data has never been loaded then it's never been
            #   accessed: no need to save it
            if self.loaded:
                t = datetime.timedelta(seconds=self.timeout * 60)
                expiration_time = self.now() + t
                if self.debug:
                    cherrypy.log('Saving session %r with expiry %s' %
                                 (self.id, expiration_time),
                                 'TOOLS.SESSIONS')
                self._save(expiration_time)
            else:
                if self.debug:
                    cherrypy.log(
                        'Skipping save of session %r (no session loaded).' %
                        self.id, 'TOOLS.SESSIONS')
        finally:
            if self.locked:
                # Always release the lock if the user didn't release it
                self.release_lock()
                if self.debug:
                    cherrypy.log('Lock released after save.', 'TOOLS.SESSIONS')

    def load(self):
        """Copy stored session data into this session instance."""
        data = self._load()
        # data is either None or a tuple (session_data, expiration_time)
        if data is None or data[1] < self.now():
            if self.debug:
                cherrypy.log('Expired session %r, flushing data.' % self.id,
                             'TOOLS.SESSIONS')
            self._data = {}
        else:
            if self.debug:
                cherrypy.log('Data loaded for session %r.' % self.id,
                             'TOOLS.SESSIONS')
            self._data = data[0]
        self.loaded = True

        # Stick the clean_thread in the class, not the instance.
        # The instances are created and destroyed per-request.
        cls = self.__class__
        if self.clean_freq and not cls.clean_thread:
            # clean_up is an instancemethod and not a classmethod,
            # so that tool config can be accessed inside the method.
            t = cherrypy.process.plugins.Monitor(
                cherrypy.engine, self.clean_up, self.clean_freq * 60,
                name='Session cleanup')
            t.subscribe()
            cls.clean_thread = t
            t.start()
            if self.debug:
                cherrypy.log('Started cleanup thread.', 'TOOLS.SESSIONS')

    def delete(self):
        """Delete stored session data."""
        self._delete()
        if self.debug:
            cherrypy.log('Deleted session %s.' % self.id,
                         'TOOLS.SESSIONS')

    # -------------------- Application accessor methods -------------------- #

    def __getitem__(self, key):
        if not self.loaded:
            self.load()
        return self._data[key]

    def __setitem__(self, key, value):
        if not self.loaded:
            self.load()
        self._data[key] = value

    def __delitem__(self, key):
        if not self.loaded:
            self.load()
        del self._data[key]

    def pop(self, key, default=missing):
        """Remove the specified key and return the corresponding value.

        If key is not found, default is returned if given, otherwise
        KeyError is raised.
        """
        if not self.loaded:
            self.load()
        if default is missing:
            return self._data.pop(key)
        else:
            return self._data.pop(key, default)

    def __contains__(self, key):
        if not self.loaded:
            self.load()
        return key in self._data

    def get(self, key, default=None):
        """D.get(k[,d]) -> D[k] if k in D, else d.

        d defaults to None.
        """
        if not self.loaded:
            self.load()
        return self._data.get(key, default)

    def update(self, d):
        """D.update(E) -> None.

        Update D from E: for k in E: D[k] = E[k].
        """
        if not self.loaded:
            self.load()
        self._data.update(d)

    def setdefault(self, key, default=None):
        """D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D."""
        if not self.loaded:
            self.load()
        return self._data.setdefault(key, default)

    def clear(self):
        """D.clear() -> None.

        Remove all items from D.
        """
        if not self.loaded:
            self.load()
        self._data.clear()

    def keys(self):
        """D.keys() -> list of D's keys."""
        if not self.loaded:
            self.load()
        return self._data.keys()

    def items(self):
        """D.items() -> list of D's (key, value) pairs, as 2-tuples."""
        if not self.loaded:
            self.load()
        return self._data.items()

    def values(self):
        """D.values() -> list of D's values."""
        if not self.loaded:
            self.load()
        return self._data.values()


class RamSession(Session):

    # Class-level objects. Don't rebind these!
    cache = {}
    locks = {}

    def clean_up(self):
        """Clean up expired sessions."""

        now = self.now()
        for _id, (data, expiration_time) in self.cache.copy().items():
            if expiration_time <= now:
                try:
                    del self.cache[_id]
                except KeyError:
                    pass
                try:
                    if self.locks[_id].acquire(blocking=False):
                        lock = self.locks.pop(_id)
                        lock.release()
                except KeyError:
                    pass

        # added to remove obsolete lock objects
        for _id in list(self.locks):
            locked = (
                _id not in self.cache
                and self.locks[_id].acquire(blocking=False)
            )
            if locked:
                lock = self.locks.pop(_id)
                lock.release()

    def _exists(self):
        return self.id in self.cache

    def _load(self):
        return self.cache.get(self.id)

    def _save(self, expiration_time):
        self.cache[self.id] = (self._data, expiration_time)

    def _delete(self):
        self.cache.pop(self.id, None)

    def acquire_lock(self):
        """Acquire an exclusive lock on the currently-loaded session data."""
        self.locked = True
        self.locks.setdefault(self.id, threading.RLock()).acquire()

    def release_lock(self):
        """Release the lock on the currently-loaded session data."""
        self.locks[self.id].release()
        self.locked = False

    def __len__(self):
        """Return the number of active sessions."""
        return len(self.cache)


class FileSession(Session):

    """Implementation of the File backend for sessions

    storage_path
        The folder where session data will be saved. Each session
        will be saved as pickle.dump(data, expiration_time) in its own file;
        the filename will be self.SESSION_PREFIX + self.id.

    lock_timeout
        A timedelta or numeric seconds indicating how long
        to block acquiring a lock. If None (default), acquiring a lock
        will block indefinitely.
    """

    SESSION_PREFIX = 'session-'
    LOCK_SUFFIX = '.lock'
    pickle_protocol = pickle.HIGHEST_PROTOCOL

    def __init__(self, id=None, **kwargs):
        # The 'storage_path' arg is required for file-based sessions.
        kwargs['storage_path'] = os.path.abspath(kwargs['storage_path'])
        kwargs.setdefault('lock_timeout', None)

        Session.__init__(self, id=id, **kwargs)

        # validate self.lock_timeout
        if isinstance(self.lock_timeout, (int, float)):
            self.lock_timeout = datetime.timedelta(seconds=self.lock_timeout)
        if not isinstance(self.lock_timeout, (datetime.timedelta, type(None))):
            raise ValueError(
                'Lock timeout must be numeric seconds or a timedelta instance.'
            )

    @classmethod
    def setup(cls, **kwargs):
        """Set up the storage system for file-based sessions.

        This should only be called once per process; this will be done
        automatically when using sessions.init (as the built-in Tool
        does).
        """
        # The 'storage_path' arg is required for file-based sessions.
        kwargs['storage_path'] = os.path.abspath(kwargs['storage_path'])

        for k, v in kwargs.items():
            setattr(cls, k, v)

    def _get_file_path(self):
        f = os.path.join(self.storage_path, self.SESSION_PREFIX + self.id)
        if not os.path.abspath(f).startswith(self.storage_path):
            raise cherrypy.HTTPError(400, 'Invalid session id in cookie.')
        return f

    def _exists(self):
        path = self._get_file_path()
        return os.path.exists(path)

    def _load(self, path=None):
        assert self.locked, ('The session load without being locked.  '
                             "Check your tools' priority levels.")
        if path is None:
            path = self._get_file_path()
        try:
            with open(path, 'rb') as f:
                return pickle.load(f)
        except (IOError, EOFError):
            e = sys.exc_info()[1]
            if self.debug:
                cherrypy.log('Error loading the session pickle: %s' %
                             e, 'TOOLS.SESSIONS')
            return None

    def _save(self, expiration_time):
        assert self.locked, ('The session was saved without being locked.  '
                             "Check your tools' priority levels.")
        with open(self._get_file_path(), 'wb') as f:
            pickle.dump((self._data, expiration_time), f, self.pickle_protocol)

    def _delete(self):
        assert self.locked, ('The session deletion without being locked.  '
                             "Check your tools' priority levels.")
        try:
            os.unlink(self._get_file_path())
        except OSError:
            pass

    def acquire_lock(self, path=None):
        """Acquire an exclusive lock on the currently-loaded session data."""
        if path is None:
            path = self._get_file_path()
        path += self.LOCK_SUFFIX
        checker = locking.LockChecker(self.id, self.lock_timeout)
        while not checker.expired():
            try:
                self.lock = zc.lockfile.LockFile(path)
            except zc.lockfile.LockError:
                time.sleep(0.1)
            else:
                break
        self.locked = True
        if self.debug:
            cherrypy.log('Lock acquired.', 'TOOLS.SESSIONS')

    def release_lock(self, path=None):
        """Release the lock on the currently-loaded session data."""
        self.lock.close()
        self.locked = False

    def clean_up(self):
        """Clean up expired sessions."""
        now = self.now()
        # Iterate over all session files in self.storage_path
        for fname in os.listdir(self.storage_path):
            have_session = (
                fname.startswith(self.SESSION_PREFIX)
                and not fname.endswith(self.LOCK_SUFFIX)
            )
            if have_session:
                # We have a session file: lock and load it and check
                #   if it's expired. If it fails, nevermind.
                path = os.path.join(self.storage_path, fname)
                self.acquire_lock(path)
                if self.debug:
                    # This is a bit of a hack, since we're calling clean_up
                    # on the first instance rather than the entire class,
                    # so depending on whether you have "debug" set on the
                    # path of the first session called, this may not run.
                    cherrypy.log('Cleanup lock acquired.', 'TOOLS.SESSIONS')

                try:
                    contents = self._load(path)
                    # _load returns None on IOError
                    if contents is not None:
                        data, expiration_time = contents
                        if expiration_time < now:
                            # Session expired: deleting it
                            os.unlink(path)
                finally:
                    self.release_lock(path)

    def __len__(self):
        """Return the number of active sessions."""
        return len([fname for fname in os.listdir(self.storage_path)
                    if (fname.startswith(self.SESSION_PREFIX) and
                        not fname.endswith(self.LOCK_SUFFIX))])


class MemcachedSession(Session):

    # The most popular memcached client for Python isn't thread-safe.
    # Wrap all .get and .set operations in a single lock.
    mc_lock = threading.RLock()

    # This is a separate set of locks per session id.
    locks = {}

    servers = ['localhost:11211']

    @classmethod
    def setup(cls, **kwargs):
        """Set up the storage system for memcached-based sessions.

        This should only be called once per process; this will be done
        automatically when using sessions.init (as the built-in Tool
        does).
        """
        for k, v in kwargs.items():
            setattr(cls, k, v)

        import memcache
        cls.cache = memcache.Client(cls.servers)

    def _exists(self):
        self.mc_lock.acquire()
        try:
            return bool(self.cache.get(self.id))
        finally:
            self.mc_lock.release()

    def _load(self):
        self.mc_lock.acquire()
        try:
            return self.cache.get(self.id)
        finally:
            self.mc_lock.release()

    def _save(self, expiration_time):
        # Send the expiration time as "Unix time" (seconds since 1/1/1970)
        td = int(time.mktime(expiration_time.timetuple()))
        self.mc_lock.acquire()
        try:
            if not self.cache.set(self.id, (self._data, expiration_time), td):
                raise AssertionError(
                    'Session data for id %r not set.' % self.id)
        finally:
            self.mc_lock.release()

    def _delete(self):
        self.cache.delete(self.id)

    def acquire_lock(self):
        """Acquire an exclusive lock on the currently-loaded session data."""
        self.locked = True
        self.locks.setdefault(self.id, threading.RLock()).acquire()
        if self.debug:
            cherrypy.log('Lock acquired.', 'TOOLS.SESSIONS')

    def release_lock(self):
        """Release the lock on the currently-loaded session data."""
        self.locks[self.id].release()
        self.locked = False

    def __len__(self):
        """Return the number of active sessions."""
        raise NotImplementedError


# Hook functions (for CherryPy tools)

def save():
    """Save any changed session data."""

    if not hasattr(cherrypy.serving, 'session'):
        return
    request = cherrypy.serving.request
    response = cherrypy.serving.response

    # Guard against running twice
    if hasattr(request, '_sessionsaved'):
        return
    request._sessionsaved = True

    if response.stream:
        # If the body is being streamed, we have to save the data
        #   *after* the response has been written out
        request.hooks.attach('on_end_request', cherrypy.session.save)
    else:
        # If the body is not being streamed, we save the data now
        # (so we can release the lock).
        if is_iterator(response.body):
            response.collapse_body()
        cherrypy.session.save()


save.failsafe = True


def close():
    """Close the session object for this request."""
    sess = getattr(cherrypy.serving, 'session', None)
    if getattr(sess, 'locked', False):
        # If the session is still locked we release the lock
        sess.release_lock()
        if sess.debug:
            cherrypy.log('Lock released on close.', 'TOOLS.SESSIONS')


close.failsafe = True
close.priority = 90


def init(storage_type=None, path=None, path_header=None, name='session_id',
         timeout=60, domain=None, secure=False, clean_freq=5,
         persistent=True, httponly=False, debug=False,
         # Py27 compat
         # *, storage_class=RamSession,
         **kwargs):
    """Initialize session object (using cookies).

    storage_class
        The Session subclass to use. Defaults to RamSession.

    storage_type
        (deprecated)
        One of 'ram', 'file', memcached'. This will be
        used to look up the corresponding class in cherrypy.lib.sessions
        globals. For example, 'file' will use the FileSession class.

    path
        The 'path' value to stick in the response cookie metadata.

    path_header
        If 'path' is None (the default), then the response
        cookie 'path' will be pulled from request.headers[path_header].

    name
        The name of the cookie.

    timeout
        The expiration timeout (in minutes) for the stored session data.
        If 'persistent' is True (the default), this is also the timeout
        for the cookie.

    domain
        The cookie domain.

    secure
        If False (the default) the cookie 'secure' value will not
        be set. If True, the cookie 'secure' value will be set (to 1).

    clean_freq (minutes)
        The poll rate for expired session cleanup.

    persistent
        If True (the default), the 'timeout' argument will be used
        to expire the cookie. If False, the cookie will not have an expiry,
        and the cookie will be a "session cookie" which expires when the
        browser is closed.

    httponly
        If False (the default) the cookie 'httponly' value will not be set.
        If True, the cookie 'httponly' value will be set (to 1).

    Any additional kwargs will be bound to the new Session instance,
    and may be specific to the storage type. See the subclass of Session
    you're using for more information.
    """

    # Py27 compat
    storage_class = kwargs.pop('storage_class', RamSession)

    request = cherrypy.serving.request

    # Guard against running twice
    if hasattr(request, '_session_init_flag'):
        return
    request._session_init_flag = True

    # Check if request came with a session ID
    id = None
    if name in request.cookie:
        id = request.cookie[name].value
        if debug:
            cherrypy.log('ID obtained from request.cookie: %r' % id,
                         'TOOLS.SESSIONS')

    first_time = not hasattr(cherrypy, 'session')

    if storage_type:
        if first_time:
            msg = 'storage_type is deprecated. Supply storage_class instead'
            cherrypy.log(msg)
        storage_class = storage_type.title() + 'Session'
        storage_class = globals()[storage_class]

    # call setup first time only
    if first_time:
        if hasattr(storage_class, 'setup'):
            storage_class.setup(**kwargs)

    # Create and attach a new Session instance to cherrypy.serving.
    # It will possess a reference to (and lock, and lazily load)
    # the requested session data.
    kwargs['timeout'] = timeout
    kwargs['clean_freq'] = clean_freq
    cherrypy.serving.session = sess = storage_class(id, **kwargs)
    sess.debug = debug

    def update_cookie(id):
        """Update the cookie every time the session id changes."""
        cherrypy.serving.response.cookie[name] = id
    sess.id_observers.append(update_cookie)

    # Create cherrypy.session which will proxy to cherrypy.serving.session
    if not hasattr(cherrypy, 'session'):
        cherrypy.session = cherrypy._ThreadLocalProxy('session')

    if persistent:
        cookie_timeout = timeout
    else:
        # See http://support.microsoft.com/kb/223799/EN-US/
        # and http://support.mozilla.com/en-US/kb/Cookies
        cookie_timeout = None
    set_response_cookie(path=path, path_header=path_header, name=name,
                        timeout=cookie_timeout, domain=domain, secure=secure,
                        httponly=httponly)


def set_response_cookie(path=None, path_header=None, name='session_id',
                        timeout=60, domain=None, secure=False, httponly=False):
    """Set a response cookie for the client.

    path
        the 'path' value to stick in the response cookie metadata.

    path_header
        if 'path' is None (the default), then the response
        cookie 'path' will be pulled from request.headers[path_header].

    name
        the name of the cookie.

    timeout
        the expiration timeout for the cookie. If 0 or other boolean
        False, no 'expires' param will be set, and the cookie will be a
        "session cookie" which expires when the browser is closed.

    domain
        the cookie domain.

    secure
        if False (the default) the cookie 'secure' value will not
        be set. If True, the cookie 'secure' value will be set (to 1).

    httponly
        If False (the default) the cookie 'httponly' value will not be set.
        If True, the cookie 'httponly' value will be set (to 1).

    """
    # Set response cookie
    cookie = cherrypy.serving.response.cookie
    cookie[name] = cherrypy.serving.session.id
    cookie[name]['path'] = (
        path or
        cherrypy.serving.request.headers.get(path_header) or
        '/'
    )

    if timeout:
        cookie[name]['max-age'] = timeout * 60
        _add_MSIE_max_age_workaround(cookie[name], timeout)
    if domain is not None:
        cookie[name]['domain'] = domain
    if secure:
        cookie[name]['secure'] = 1
    if httponly:
        if not cookie[name].isReservedKey('httponly'):
            raise ValueError('The httponly cookie token is not supported.')
        cookie[name]['httponly'] = 1


def _add_MSIE_max_age_workaround(cookie, timeout):
    """
    We'd like to use the "max-age" param as indicated in
    http://www.faqs.org/rfcs/rfc2109.html but IE doesn't
    save it to disk and the session is lost if people close
    the browser. So we have to use the old "expires" ... sigh ...
    """
    expires = time.time() + timeout * 60
    cookie['expires'] = httputil.HTTPDate(expires)


def expire():
    """Expire the current session cookie."""
    name = cherrypy.serving.request.config.get(
        'tools.sessions.name', 'session_id')
    one_year = 60 * 60 * 24 * 365
    e = time.time() - one_year
    cherrypy.serving.response.cookie[name]['expires'] = httputil.HTTPDate(e)
    cherrypy.serving.response.cookie[name].pop('max-age', None)



"""Blocking and non-blocking HTTP client interfaces.

This module defines a common interface shared by two implementations,
``simple_httpclient`` and ``curl_httpclient``.  Applications may either
instantiate their chosen implementation class directly or use the
`AsyncHTTPClient` class from this module, which selects an implementation
that can be overridden with the `AsyncHTTPClient.configure` method.

The default implementation is ``simple_httpclient``, and this is expected
to be suitable for most users' needs.  However, some applications may wish
to switch to ``curl_httpclient`` for reasons such as the following:

* ``curl_httpclient`` has some features not found in ``simple_httpclient``,
  including support for HTTP proxies and the ability to use a specified
  network interface.

* ``curl_httpclient`` is more likely to be compatible with sites that are
  not-quite-compliant with the HTTP spec, or sites that use little-exercised
  features of HTTP.

* ``curl_httpclient`` is faster.

Note that if you are using ``curl_httpclient``, it is highly
recommended that you use a recent version of ``libcurl`` and
``pycurl``.  Currently the minimum supported version of libcurl is
7.22.0, and the minimum version of pycurl is 7.18.2.  It is highly
recommended that your ``libcurl`` installation is built with
asynchronous DNS resolver (threaded or c-ares), otherwise you may
encounter various problems with request timeouts (for more
information, see
http://curl.haxx.se/libcurl/c/curl_easy_setopt.html#CURLOPTCONNECTTIMEOUTMS
and comments in curl_httpclient.py).

To select ``curl_httpclient``, call `AsyncHTTPClient.configure` at startup::

    AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")
"""

import datetime
import functools
from io import BytesIO
import ssl
import time
import weakref

from tornado.concurrent import (
    Future,
    future_set_result_unless_cancelled,
    future_set_exception_unless_cancelled,
)
from tornado.escape import utf8, native_str
from tornado import gen, httputil
from tornado.ioloop import IOLoop
from tornado.util import Configurable

from typing import Type, Any, Union, Dict, Callable, Optional, cast


class HTTPClient:
    """A blocking HTTP client.

    This interface is provided to make it easier to share code between
    synchronous and asynchronous applications. Applications that are
    running an `.IOLoop` must use `AsyncHTTPClient` instead.

    Typical usage looks like this::

        http_client = httpclient.HTTPClient()
        try:
            response = http_client.fetch("http://www.google.com/")
            print(response.body)
        except httpclient.HTTPError as e:
            # HTTPError is raised for non-200 responses; the response
            # can be found in e.response.
            print("Error: " + str(e))
        except Exception as e:
            # Other errors are possible, such as IOError.
            print("Error: " + str(e))
        http_client.close()

    .. versionchanged:: 5.0

       Due to limitations in `asyncio`, it is no longer possible to
       use the synchronous ``HTTPClient`` while an `.IOLoop` is running.
       Use `AsyncHTTPClient` instead.

    """

    def __init__(
        self,
        async_client_class: "Optional[Type[AsyncHTTPClient]]" = None,
        **kwargs: Any,
    ) -> None:
        # Initialize self._closed at the beginning of the constructor
        # so that an exception raised here doesn't lead to confusing
        # failures in __del__.
        self._closed = True
        self._io_loop = IOLoop(make_current=False)
        if async_client_class is None:
            async_client_class = AsyncHTTPClient

        # Create the client while our IOLoop is "current", without
        # clobbering the thread's real current IOLoop (if any).
        async def make_client() -> "AsyncHTTPClient":
            await gen.sleep(0)
            assert async_client_class is not None
            return async_client_class(**kwargs)

        self._async_client = self._io_loop.run_sync(make_client)
        self._closed = False

    def __del__(self) -> None:
        self.close()

    def close(self) -> None:
        """Closes the HTTPClient, freeing any resources used."""
        if not self._closed:
            self._async_client.close()
            self._io_loop.close()
            self._closed = True

    def fetch(
        self, request: Union["HTTPRequest", str], **kwargs: Any
    ) -> "HTTPResponse":
        """Executes a request, returning an `HTTPResponse`.

        The request may be either a string URL or an `HTTPRequest` object.
        If it is a string, we construct an `HTTPRequest` using any additional
        kwargs: ``HTTPRequest(request, **kwargs)``

        If an error occurs during the fetch, we raise an `HTTPError` unless
        the ``raise_error`` keyword argument is set to False.
        """
        response = self._io_loop.run_sync(
            functools.partial(self._async_client.fetch, request, **kwargs)
        )
        return response


class AsyncHTTPClient(Configurable):
    """An non-blocking HTTP client.

    Example usage::

        async def f():
            http_client = AsyncHTTPClient()
            try:
                response = await http_client.fetch("http://www.google.com")
            except Exception as e:
                print("Error: %s" % e)
            else:
                print(response.body)

    The constructor for this class is magic in several respects: It
    actually creates an instance of an implementation-specific
    subclass, and instances are reused as a kind of pseudo-singleton
    (one per `.IOLoop`). The keyword argument ``force_instance=True``
    can be used to suppress this singleton behavior. Unless
    ``force_instance=True`` is used, no arguments should be passed to
    the `AsyncHTTPClient` constructor. The implementation subclass as
    well as arguments to its constructor can be set with the static
    method `configure()`

    All `AsyncHTTPClient` implementations support a ``defaults``
    keyword argument, which can be used to set default values for
    `HTTPRequest` attributes.  For example::

        AsyncHTTPClient.configure(
            None, defaults=dict(user_agent="MyUserAgent"))
        # or with force_instance:
        client = AsyncHTTPClient(force_instance=True,
            defaults=dict(user_agent="MyUserAgent"))

    .. versionchanged:: 5.0
       The ``io_loop`` argument (deprecated since version 4.1) has been removed.

    """

    _instance_cache = None  # type: Dict[IOLoop, AsyncHTTPClient]

    @classmethod
    def configurable_base(cls) -> Type[Configurable]:
        return AsyncHTTPClient

    @classmethod
    def configurable_default(cls) -> Type[Configurable]:
        from tornado.simple_httpclient import SimpleAsyncHTTPClient

        return SimpleAsyncHTTPClient

    @classmethod
    def _async_clients(cls) -> Dict[IOLoop, "AsyncHTTPClient"]:
        attr_name = "_async_client_dict_" + cls.__name__
        if not hasattr(cls, attr_name):
            setattr(cls, attr_name, weakref.WeakKeyDictionary())
        return getattr(cls, attr_name)

    def __new__(cls, force_instance: bool = False, **kwargs: Any) -> "AsyncHTTPClient":
        io_loop = IOLoop.current()
        if force_instance:
            instance_cache = None
        else:
            instance_cache = cls._async_clients()
        if instance_cache is not None and io_loop in instance_cache:
            return instance_cache[io_loop]
        instance = super().__new__(cls, **kwargs)  # type: ignore
        # Make sure the instance knows which cache to remove itself from.
        # It can't simply call _async_clients() because we may be in
        # __new__(AsyncHTTPClient) but instance.__class__ may be
        # SimpleAsyncHTTPClient.
        instance._instance_cache = instance_cache
        if instance_cache is not None:
            instance_cache[instance.io_loop] = instance
        return instance

    def initialize(self, defaults: Optional[Dict[str, Any]] = None) -> None:
        self.io_loop = IOLoop.current()
        self.defaults = dict(HTTPRequest._DEFAULTS)
        if defaults is not None:
            self.defaults.update(defaults)
        self._closed = False

    def close(self) -> None:
        """Destroys this HTTP client, freeing any file descriptors used.

        This method is **not needed in normal use** due to the way
        that `AsyncHTTPClient` objects are transparently reused.
        ``close()`` is generally only necessary when either the
        `.IOLoop` is also being closed, or the ``force_instance=True``
        argument was used when creating the `AsyncHTTPClient`.

        No other methods may be called on the `AsyncHTTPClient` after
        ``close()``.

        """
        if self._closed:
            return
        self._closed = True
        if self._instance_cache is not None:
            cached_val = self._instance_cache.pop(self.io_loop, None)
            # If there's an object other than self in the instance
            # cache for our IOLoop, something has gotten mixed up. A
            # value of None appears to be possible when this is called
            # from a destructor (HTTPClient.__del__) as the weakref
            # gets cleared before the destructor runs.
            if cached_val is not None and cached_val is not self:
                raise RuntimeError("inconsistent AsyncHTTPClient cache")

    def fetch(
        self,
        request: Union[str, "HTTPRequest"],
        raise_error: bool = True,
        **kwargs: Any,
    ) -> "Future[HTTPResponse]":
        """Executes a request, asynchronously returning an `HTTPResponse`.

        The request may be either a string URL or an `HTTPRequest` object.
        If it is a string, we construct an `HTTPRequest` using any additional
        kwargs: ``HTTPRequest(request, **kwargs)``

        This method returns a `.Future` whose result is an
        `HTTPResponse`. By default, the ``Future`` will raise an
        `HTTPError` if the request returned a non-200 response code
        (other errors may also be raised if the server could not be
        contacted). Instead, if ``raise_error`` is set to False, the
        response will always be returned regardless of the response
        code.

        If a ``callback`` is given, it will be invoked with the `HTTPResponse`.
        In the callback interface, `HTTPError` is not automatically raised.
        Instead, you must check the response's ``error`` attribute or
        call its `~HTTPResponse.rethrow` method.

        .. versionchanged:: 6.0

           The ``callback`` argument was removed. Use the returned
           `.Future` instead.

           The ``raise_error=False`` argument only affects the
           `HTTPError` raised when a non-200 response code is used,
           instead of suppressing all errors.
        """
        if self._closed:
            raise RuntimeError("fetch() called on closed AsyncHTTPClient")
        if not isinstance(request, HTTPRequest):
            request = HTTPRequest(url=request, **kwargs)
        else:
            if kwargs:
                raise ValueError(
                    "kwargs can't be used if request is an HTTPRequest object"
                )
        # We may modify this (to add Host, Accept-Encoding, etc),
        # so make sure we don't modify the caller's object.  This is also
        # where normal dicts get converted to HTTPHeaders objects.
        request.headers = httputil.HTTPHeaders(request.headers)
        request_proxy = _RequestProxy(request, self.defaults)
        future = Future()  # type: Future[HTTPResponse]

        def handle_response(response: "HTTPResponse") -> None:
            if response.error:
                if raise_error or not response._error_is_response_code:
                    future_set_exception_unless_cancelled(future, response.error)
                    return
            future_set_result_unless_cancelled(future, response)

        self.fetch_impl(cast(HTTPRequest, request_proxy), handle_response)
        return future

    def fetch_impl(
        self, request: "HTTPRequest", callback: Callable[["HTTPResponse"], None]
    ) -> None:
        raise NotImplementedError()

    @classmethod
    def configure(
        cls, impl: "Union[None, str, Type[Configurable]]", **kwargs: Any
    ) -> None:
        """Configures the `AsyncHTTPClient` subclass to use.

        ``AsyncHTTPClient()`` actually creates an instance of a subclass.
        This method may be called with either a class object or the
        fully-qualified name of such a class (or ``None`` to use the default,
        ``SimpleAsyncHTTPClient``)

        If additional keyword arguments are given, they will be passed
        to the constructor of each subclass instance created.  The
        keyword argument ``max_clients`` determines the maximum number
        of simultaneous `~AsyncHTTPClient.fetch()` operations that can
        execute in parallel on each `.IOLoop`.  Additional arguments
        may be supported depending on the implementation class in use.

        Example::

           AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")
        """
        super().configure(impl, **kwargs)


class HTTPRequest:
    """HTTP client request object."""

    _headers = None  # type: Union[Dict[str, str], httputil.HTTPHeaders]

    # Default values for HTTPRequest parameters.
    # Merged with the values on the request object by AsyncHTTPClient
    # implementations.
    _DEFAULTS = dict(
        connect_timeout=20.0,
        request_timeout=20.0,
        follow_redirects=True,
        max_redirects=5,
        decompress_response=True,
        proxy_password="",
        allow_nonstandard_methods=False,
        validate_cert=True,
    )

    def __init__(
        self,
        url: str,
        method: str = "GET",
        headers: Optional[Union[Dict[str, str], httputil.HTTPHeaders]] = None,
        body: Optional[Union[bytes, str]] = None,
        auth_username: Optional[str] = None,
        auth_password: Optional[str] = None,
        auth_mode: Optional[str] = None,
        connect_timeout: Optional[float] = None,
        request_timeout: Optional[float] = None,
        if_modified_since: Optional[Union[float, datetime.datetime]] = None,
        follow_redirects: Optional[bool] = None,
        max_redirects: Optional[int] = None,
        user_agent: Optional[str] = None,
        use_gzip: Optional[bool] = None,
        network_interface: Optional[str] = None,
        streaming_callback: Optional[Callable[[bytes], None]] = None,
        header_callback: Optional[Callable[[str], None]] = None,
        prepare_curl_callback: Optional[Callable[[Any], None]] = None,
        proxy_host: Optional[str] = None,
        proxy_port: Optional[int] = None,
        proxy_username: Optional[str] = None,
        proxy_password: Optional[str] = None,
        proxy_auth_mode: Optional[str] = None,
        allow_nonstandard_methods: Optional[bool] = None,
        validate_cert: Optional[bool] = None,
        ca_certs: Optional[str] = None,
        allow_ipv6: Optional[bool] = None,
        client_key: Optional[str] = None,
        client_cert: Optional[str] = None,
        body_producer: Optional[
            Callable[[Callable[[bytes], None]], "Future[None]"]
        ] = None,
        expect_100_continue: bool = False,
        decompress_response: Optional[bool] = None,
        ssl_options: Optional[Union[Dict[str, Any], ssl.SSLContext]] = None,
    ) -> None:
        r"""All parameters except ``url`` are optional.

        :arg str url: URL to fetch
        :arg str method: HTTP method, e.g. "GET" or "POST"
        :arg headers: Additional HTTP headers to pass on the request
        :type headers: `~tornado.httputil.HTTPHeaders` or `dict`
        :arg body: HTTP request body as a string (byte or unicode; if unicode
           the utf-8 encoding will be used)
        :type body: `str` or `bytes`
        :arg collections.abc.Callable body_producer: Callable used for
           lazy/asynchronous request bodies.
           It is called with one argument, a ``write`` function, and should
           return a `.Future`.  It should call the write function with new
           data as it becomes available.  The write function returns a
           `.Future` which can be used for flow control.
           Only one of ``body`` and ``body_producer`` may
           be specified.  ``body_producer`` is not supported on
           ``curl_httpclient``.  When using ``body_producer`` it is recommended
           to pass a ``Content-Length`` in the headers as otherwise chunked
           encoding will be used, and many servers do not support chunked
           encoding on requests.  New in Tornado 4.0
        :arg str auth_username: Username for HTTP authentication
        :arg str auth_password: Password for HTTP authentication
        :arg str auth_mode: Authentication mode; default is "basic".
           Allowed values are implementation-defined; ``curl_httpclient``
           supports "basic" and "digest"; ``simple_httpclient`` only supports
           "basic"
        :arg float connect_timeout: Timeout for initial connection in seconds,
           default 20 seconds (0 means no timeout)
        :arg float request_timeout: Timeout for entire request in seconds,
           default 20 seconds (0 means no timeout)
        :arg if_modified_since: Timestamp for ``If-Modified-Since`` header
        :type if_modified_since: `datetime` or `float`
        :arg bool follow_redirects: Should redirects be followed automatically
           or return the 3xx response? Default True.
        :arg int max_redirects: Limit for ``follow_redirects``, default 5.
        :arg str user_agent: String to send as ``User-Agent`` header
        :arg bool decompress_response: Request a compressed response from
           the server and decompress it after downloading.  Default is True.
           New in Tornado 4.0.
        :arg bool use_gzip: Deprecated alias for ``decompress_response``
           since Tornado 4.0.
        :arg str network_interface: Network interface or source IP to use for request.
           See ``curl_httpclient`` note below.
        :arg collections.abc.Callable streaming_callback: If set, ``streaming_callback`` will
           be run with each chunk of data as it is received, and
           ``HTTPResponse.body`` and ``HTTPResponse.buffer`` will be empty in
           the final response.
        :arg collections.abc.Callable header_callback: If set, ``header_callback`` will
           be run with each header line as it is received (including the
           first line, e.g. ``HTTP/1.0 200 OK\r\n``, and a final line
           containing only ``\r\n``.  All lines include the trailing newline
           characters).  ``HTTPResponse.headers`` will be empty in the final
           response.  This is most useful in conjunction with
           ``streaming_callback``, because it's the only way to get access to
           header data while the request is in progress.
        :arg collections.abc.Callable prepare_curl_callback: If set, will be called with
           a ``pycurl.Curl`` object to allow the application to make additional
           ``setopt`` calls.
        :arg str proxy_host: HTTP proxy hostname.  To use proxies,
           ``proxy_host`` and ``proxy_port`` must be set; ``proxy_username``,
           ``proxy_pass`` and ``proxy_auth_mode`` are optional.  Proxies are
           currently only supported with ``curl_httpclient``.
        :arg int proxy_port: HTTP proxy port
        :arg str proxy_username: HTTP proxy username
        :arg str proxy_password: HTTP proxy password
        :arg str proxy_auth_mode: HTTP proxy Authentication mode;
           default is "basic". supports "basic" and "digest"
        :arg bool allow_nonstandard_methods: Allow unknown values for ``method``
           argument? Default is False.
        :arg bool validate_cert: For HTTPS requests, validate the server's
           certificate? Default is True.
        :arg str ca_certs: filename of CA certificates in PEM format,
           or None to use defaults.  See note below when used with
           ``curl_httpclient``.
        :arg str client_key: Filename for client SSL key, if any.  See
           note below when used with ``curl_httpclient``.
        :arg str client_cert: Filename for client SSL certificate, if any.
           See note below when used with ``curl_httpclient``.
        :arg ssl.SSLContext ssl_options: `ssl.SSLContext` object for use in
           ``simple_httpclient`` (unsupported by ``curl_httpclient``).
           Overrides ``validate_cert``, ``ca_certs``, ``client_key``,
           and ``client_cert``.
        :arg bool allow_ipv6: Use IPv6 when available?  Default is True.
        :arg bool expect_100_continue: If true, send the
           ``Expect: 100-continue`` header and wait for a continue response
           before sending the request body.  Only supported with
           ``simple_httpclient``.

        .. note::

            When using ``curl_httpclient`` certain options may be
            inherited by subsequent fetches because ``pycurl`` does
            not allow them to be cleanly reset.  This applies to the
            ``ca_certs``, ``client_key``, ``client_cert``, and
            ``network_interface`` arguments.  If you use these
            options, you should pass them on every request (you don't
            have to always use the same values, but it's not possible
            to mix requests that specify these options with ones that
            use the defaults).

        .. versionadded:: 3.1
           The ``auth_mode`` argument.

        .. versionadded:: 4.0
           The ``body_producer`` and ``expect_100_continue`` arguments.

        .. versionadded:: 4.2
           The ``ssl_options`` argument.

        .. versionadded:: 4.5
           The ``proxy_auth_mode`` argument.
        """
        # Note that some of these attributes go through property setters
        # defined below.
        self.headers = headers  # type: ignore
        if if_modified_since:
            self.headers["If-Modified-Since"] = httputil.format_timestamp(
                if_modified_since
            )
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.proxy_username = proxy_username
        self.proxy_password = proxy_password
        self.proxy_auth_mode = proxy_auth_mode
        self.url = url
        self.method = method
        self.body = body  # type: ignore
        self.body_producer = body_producer
        self.auth_username = auth_username
        self.auth_password = auth_password
        self.auth_mode = auth_mode
        self.connect_timeout = connect_timeout
        self.request_timeout = request_timeout
        self.follow_redirects = follow_redirects
        self.max_redirects = max_redirects
        self.user_agent = user_agent
        if decompress_response is not None:
            self.decompress_response = decompress_response  # type: Optional[bool]
        else:
            self.decompress_response = use_gzip
        self.network_interface = network_interface
        self.streaming_callback = streaming_callback
        self.header_callback = header_callback
        self.prepare_curl_callback = prepare_curl_callback
        self.allow_nonstandard_methods = allow_nonstandard_methods
        self.validate_cert = validate_cert
        self.ca_certs = ca_certs
        self.allow_ipv6 = allow_ipv6
        self.client_key = client_key
        self.client_cert = client_cert
        self.ssl_options = ssl_options
        self.expect_100_continue = expect_100_continue
        self.start_time = time.time()

    @property
    def headers(self) -> httputil.HTTPHeaders:
        # TODO: headers may actually be a plain dict until fairly late in
        # the process (AsyncHTTPClient.fetch), but practically speaking,
        # whenever the property is used they're already HTTPHeaders.
        return self._headers  # type: ignore

    @headers.setter
    def headers(self, value: Union[Dict[str, str], httputil.HTTPHeaders]) -> None:
        if value is None:
            self._headers = httputil.HTTPHeaders()
        else:
            self._headers = value  # type: ignore

    @property
    def body(self) -> bytes:
        return self._body

    @body.setter
    def body(self, value: Union[bytes, str]) -> None:
        self._body = utf8(value)


class HTTPResponse:
    """HTTP Response object.

    Attributes:

    * ``request``: HTTPRequest object

    * ``code``: numeric HTTP status code, e.g. 200 or 404

    * ``reason``: human-readable reason phrase describing the status code

    * ``headers``: `tornado.httputil.HTTPHeaders` object

    * ``effective_url``: final location of the resource after following any
      redirects

    * ``buffer``: ``cStringIO`` object for response body

    * ``body``: response body as bytes (created on demand from ``self.buffer``)

    * ``error``: Exception object, if any

    * ``request_time``: seconds from request start to finish. Includes all
      network operations from DNS resolution to receiving the last byte of
      data. Does not include time spent in the queue (due to the
      ``max_clients`` option). If redirects were followed, only includes
      the final request.

    * ``start_time``: Time at which the HTTP operation started, based on
      `time.time` (not the monotonic clock used by `.IOLoop.time`). May
      be ``None`` if the request timed out while in the queue.

    * ``time_info``: dictionary of diagnostic timing information from the
      request. Available data are subject to change, but currently uses timings
      available from http://curl.haxx.se/libcurl/c/curl_easy_getinfo.html,
      plus ``queue``, which is the delay (if any) introduced by waiting for
      a slot under `AsyncHTTPClient`'s ``max_clients`` setting.

    .. versionadded:: 5.1

       Added the ``start_time`` attribute.

    .. versionchanged:: 5.1

       The ``request_time`` attribute previously included time spent in the queue
       for ``simple_httpclient``, but not in ``curl_httpclient``. Now queueing time
       is excluded in both implementations. ``request_time`` is now more accurate for
       ``curl_httpclient`` because it uses a monotonic clock when available.
    """

    # I'm not sure why these don't get type-inferred from the references in __init__.
    error = None  # type: Optional[BaseException]
    _error_is_response_code = False
    request = None  # type: HTTPRequest

    def __init__(
        self,
        request: HTTPRequest,
        code: int,
        headers: Optional[httputil.HTTPHeaders] = None,
        buffer: Optional[BytesIO] = None,
        effective_url: Optional[str] = None,
        error: Optional[BaseException] = None,
        request_time: Optional[float] = None,
        time_info: Optional[Dict[str, float]] = None,
        reason: Optional[str] = None,
        start_time: Optional[float] = None,
    ) -> None:
        if isinstance(request, _RequestProxy):
            self.request = request.request
        else:
            self.request = request
        self.code = code
        self.reason = reason or httputil.responses.get(code, "Unknown")
        if headers is not None:
            self.headers = headers
        else:
            self.headers = httputil.HTTPHeaders()
        self.buffer = buffer
        self._body = None  # type: Optional[bytes]
        if effective_url is None:
            self.effective_url = request.url
        else:
            self.effective_url = effective_url
        self._error_is_response_code = False
        if error is None:
            if self.code < 200 or self.code >= 300:
                self._error_is_response_code = True
                self.error = HTTPError(self.code, message=self.reason, response=self)
            else:
                self.error = None
        else:
            self.error = error
        self.start_time = start_time
        self.request_time = request_time
        self.time_info = time_info or {}

    @property
    def body(self) -> bytes:
        if self.buffer is None:
            return b""
        elif self._body is None:
            self._body = self.buffer.getvalue()

        return self._body

    def rethrow(self) -> None:
        """If there was an error on the request, raise an `HTTPError`."""
        if self.error:
            raise self.error

    def __repr__(self) -> str:
        args = ",".join("%s=%r" % i for i in sorted(self.__dict__.items()))
        return f"{self.__class__.__name__}({args})"


class HTTPClientError(Exception):
    """Exception thrown for an unsuccessful HTTP request.

    Attributes:

    * ``code`` - HTTP error integer error code, e.g. 404.  Error code 599 is
      used when no HTTP response was received, e.g. for a timeout.

    * ``response`` - `HTTPResponse` object, if any.

    Note that if ``follow_redirects`` is False, redirects become HTTPErrors,
    and you can look at ``error.response.headers['Location']`` to see the
    destination of the redirect.

    .. versionchanged:: 5.1

       Renamed from ``HTTPError`` to ``HTTPClientError`` to avoid collisions with
       `tornado.web.HTTPError`. The name ``tornado.httpclient.HTTPError`` remains
       as an alias.
    """

    def __init__(
        self,
        code: int,
        message: Optional[str] = None,
        response: Optional[HTTPResponse] = None,
    ) -> None:
        self.code = code
        self.message = message or httputil.responses.get(code, "Unknown")
        self.response = response
        super().__init__(code, message, response)

    def __str__(self) -> str:
        return "HTTP %d: %s" % (self.code, self.message)

    # There is a cyclic reference between self and self.response,
    # which breaks the default __repr__ implementation.
    # (especially on pypy, which doesn't have the same recursion
    # detection as cpython).
    __repr__ = __str__


HTTPError = HTTPClientError


class _RequestProxy:
    """Combines an object with a dictionary of defaults.

    Used internally by AsyncHTTPClient implementations.
    """

    def __init__(
        self, request: HTTPRequest, defaults: Optional[Dict[str, Any]]
    ) -> None:
        self.request = request
        self.defaults = defaults

    def __getattr__(self, name: str) -> Any:
        request_attr = getattr(self.request, name)
        if request_attr is not None:
            return request_attr
        elif self.defaults is not None:
            return self.defaults.get(name, None)
        else:
            return None


def main() -> None:
    from tornado.options import define, options, parse_command_line

    define("print_headers", type=bool, default=False)
    define("print_body", type=bool, default=True)
    define("follow_redirects", type=bool, default=True)
    define("validate_cert", type=bool, default=True)
    define("proxy_host", type=str)
    define("proxy_port", type=int)
    args = parse_command_line()
    client = HTTPClient()
    for arg in args:
        try:
            response = client.fetch(
                arg,
                follow_redirects=options.follow_redirects,
                validate_cert=options.validate_cert,
                proxy_host=options.proxy_host,
                proxy_port=options.proxy_port,
            )
        except HTTPError as e:
            if e.response is not None:
                response = e.response
            else:
                raise
        if options.print_headers:
            print(response.headers)
        if options.print_body:
            print(native_str(response.body))
    client.close()


if __name__ == "__main__":
    main()



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import abc
import enum
import logging
import sys
from collections.abc import Callable, Sequence
from os import environ, linesep
from typing import IO

from typing_extensions import deprecated

from opentelemetry.context import (
    _ON_EMIT_RECURSION_COUNT_KEY,
    _SUPPRESS_INSTRUMENTATION_KEY,
    attach,
    detach,
    get_value,
    set_value,
)
from opentelemetry.metrics import MeterProvider, get_meter_provider
from opentelemetry.sdk._logs import (
    LogRecordProcessor,
    ReadableLogRecord,
    ReadWriteLogRecord,
)
from opentelemetry.sdk._shared_internal import (
    BatchProcessor,
    DuplicateFilter,
)
from opentelemetry.sdk._shared_internal._processor_metrics import (
    create_processor_metrics,
)
from opentelemetry.sdk.environment_variables import (
    OTEL_BLRP_EXPORT_TIMEOUT,
    OTEL_BLRP_MAX_EXPORT_BATCH_SIZE,
    OTEL_BLRP_MAX_QUEUE_SIZE,
    OTEL_BLRP_SCHEDULE_DELAY,
    OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED,
)
from opentelemetry.sdk.environment_variables._internal import (
    parse_boolean_environment_variable,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.semconv._incubating.attributes.otel_attributes import (
    OtelComponentTypeValues,
)

_DEFAULT_SCHEDULE_DELAY_MILLIS = 1000
_DEFAULT_MAX_EXPORT_BATCH_SIZE = 512
_DEFAULT_EXPORT_TIMEOUT_MILLIS = 30000
_DEFAULT_MAX_QUEUE_SIZE = 2048
_ENV_VAR_INT_VALUE_ERROR_MESSAGE = (
    "Unable to parse value for %s as integer. Defaulting to %s."
)
_logger = logging.getLogger(__name__)
_logger.addFilter(DuplicateFilter())

_propagate_false_logger = logging.getLogger(__name__ + ".propagate.false")
_propagate_false_logger.propagate = False


class LogRecordExportResult(enum.Enum):
    SUCCESS = 0
    FAILURE = 1


@deprecated(
    "Use LogRecordExportResult. Since logs are not stable yet this WILL be removed in future releases."
)
class LogExportResult(enum.Enum):
    SUCCESS = 0
    FAILURE = 1


class LogRecordExporter(abc.ABC):
    """Interface for exporting logs.

    Interface to be implemented by services that want to export logs received
    in their own format.

    To export data this MUST be registered to the :class:`opentelemetry.sdk._logs.Logger`
    using a log processor.

    Important
    ---------
    The ``export()`` method may raise exceptions (e.g., network errors,
    timeouts, serialization errors). It is the responsibility of the
    ``LogRecordProcessor`` calling this exporter to handle these exceptions
    appropriately to prevent application crashes. See ``LogRecordProcessor``
    for guidance on implementing proper error handling.
    """

    @abc.abstractmethod
    def export(
        self, batch: Sequence[ReadableLogRecord]
    ) -> LogRecordExportResult:
        """Exports a batch of logs.

        Args:
            batch: The list of ``ReadableLogRecord`` objects to be exported.

        Returns:
            The result of the export.

        Raises:
            Exception: This method may raise exceptions on network errors,
                timeouts, or other failures. Callers (i.e., log processors)
                should handle these exceptions to comply with OpenTelemetry
                error handling principles.
        """

    @abc.abstractmethod
    def shutdown(self):
        """Shuts down the exporter.

        Called when the SDK is shut down.
        """


@deprecated(
    "Use LogRecordExporter. Since logs are not stable yet this WILL be removed in future releases."
)
class LogExporter(LogRecordExporter):
    pass


class ConsoleLogRecordExporter(LogRecordExporter):
    """Implementation of :class:`LogRecordExporter` that prints log records to the
    console.

    This class can be used for diagnostic purposes. It prints the exported
    log records to the console STDOUT.
    """

    def __init__(
        self,
        out: IO = sys.stdout,
        formatter: Callable[[ReadableLogRecord], str] = lambda record: (
            record.to_json() + linesep
        ),
    ):
        self.out = out
        self.formatter = formatter

    def export(self, batch: Sequence[ReadableLogRecord]):
        for log_record in batch:
            self.out.write(self.formatter(log_record))
        self.out.flush()
        return LogRecordExportResult.SUCCESS

    def shutdown(self):
        pass


@deprecated(
    "Use ConsoleLogRecordExporter. Since logs are not stable yet this WILL be removed in future releases."
)
class ConsoleLogExporter(ConsoleLogRecordExporter):
    pass


class SimpleLogRecordProcessor(LogRecordProcessor):
    """Implementation of LogRecordProcessor that exports logs synchronously.

    This processor passes received logs directly to the configured
    ``LogRecordExporter`` as soon as they are emitted.

    This class serves as a reference implementation for custom log processors,
    demonstrating proper error handling. Note how the ``on_emit`` method wraps
    the exporter call in a try/except block to prevent exceptions from
    propagating to the application.
    """

    def __init__(
        self,
        exporter: LogRecordExporter,
        *,
        meter_provider: MeterProvider | None = None,
    ):
        self._exporter = exporter
        self._shutdown = False
        self._metrics = create_processor_metrics(
            "logs",
            OtelComponentTypeValues.SIMPLE_LOG_PROCESSOR,
            meter_provider or get_meter_provider(),
            enabled=parse_boolean_environment_variable(
                OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED
            ),
        )

    def on_emit(self, log_record: ReadWriteLogRecord):
        # Prevent entering a recursive loop.
        cnt = get_value(_ON_EMIT_RECURSION_COUNT_KEY) or 0
        # Recursive depth of 3 is sort of arbitrary. It's possible that an Exporter.export call
        # emits a log which returns us to this function, but when we call Exporter.export again the log
        # is no longer emitted and we exit this recursive loop naturally, a depth of >3 allows 3
        # recursive log calls but exits after because it's likely endless.
        if cnt > 3:  # pyright: ignore[reportOperatorIssue]
            _propagate_false_logger.warning(
                "SimpleLogRecordProcessor.on_emit has entered a recursive loop. Dropping log and exiting the loop."
            )
            return
        token = attach(
            set_value(
                _SUPPRESS_INSTRUMENTATION_KEY,
                True,
                set_value(_ON_EMIT_RECURSION_COUNT_KEY, cnt + 1),  # pyright: ignore[reportOperatorIssue]
            )
        )
        error: Exception | None = None
        try:
            if self._shutdown:
                _logger.warning("Processor is already shutdown, ignoring call")
                return
            # Convert ReadWriteLogRecord to ReadableLogRecord before exporting
            # Note: resource should not be None at this point as it's set during Logger.emit()
            resource = (
                log_record.resource
                if log_record.resource is not None
                else Resource.create({})
            )
            readable_log_record = ReadableLogRecord(
                log_record=log_record.log_record,
                resource=resource,
                instrumentation_scope=log_record.instrumentation_scope,
                limits=log_record.limits,
            )
            self._exporter.export((readable_log_record,))
        except Exception as err:  # pylint: disable=broad-exception-caught
            error = err
            _logger.exception("Exception while exporting logs.")
        finally:
            self._metrics.finish_items(1, error)
            detach(token)

    def shutdown(self):
        self._shutdown = True
        self._exporter.shutdown()

    def force_flush(self, timeout_millis: int = 30000) -> bool:  # pylint: disable=no-self-use
        return True


class BatchLogRecordProcessor(LogRecordProcessor):
    """This is an implementation of LogRecordProcessor which creates batches of
    received logs and sends them to the configured LogRecordExporter.

    `BatchLogRecordProcessor` is configurable with the following environment
    variables which correspond to constructor parameters:

    - :envvar:`OTEL_BLRP_SCHEDULE_DELAY`
    - :envvar:`OTEL_BLRP_MAX_QUEUE_SIZE`
    - :envvar:`OTEL_BLRP_MAX_EXPORT_BATCH_SIZE`
    - :envvar:`OTEL_BLRP_EXPORT_TIMEOUT`

    All the logic for emitting logs, shutting down etc. resides in the BatchProcessor class.
    """

    def __init__(
        self,
        exporter: LogRecordExporter,
        schedule_delay_millis: float | None = None,
        max_export_batch_size: int | None = None,
        export_timeout_millis: float | None = None,
        max_queue_size: int | None = None,
        *,
        meter_provider: MeterProvider | None = None,
    ):
        if max_queue_size is None:
            max_queue_size = BatchLogRecordProcessor._default_max_queue_size()

        if schedule_delay_millis is None:
            schedule_delay_millis = (
                BatchLogRecordProcessor._default_schedule_delay_millis()
            )

        if max_export_batch_size is None:
            max_export_batch_size = (
                BatchLogRecordProcessor._default_max_export_batch_size()
            )
        # Not used. No way currently to pass timeout to export.
        if export_timeout_millis is None:
            export_timeout_millis = (
                BatchLogRecordProcessor._default_export_timeout_millis()
            )

        BatchLogRecordProcessor._validate_arguments(
            max_queue_size, schedule_delay_millis, max_export_batch_size
        )
        # Initializes BatchProcessor
        self._batch_processor = BatchProcessor(
            exporter,
            schedule_delay_millis,
            max_export_batch_size,
            export_timeout_millis,
            max_queue_size,
            "Log",
            create_processor_metrics(
                "logs",
                OtelComponentTypeValues.BATCHING_LOG_PROCESSOR,
                meter_provider or get_meter_provider(),
                capacity=max_queue_size,
                enabled=parse_boolean_environment_variable(
                    OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED
                ),
            ),
        )

    def on_emit(self, log_record: ReadWriteLogRecord) -> None:
        # Convert ReadWriteLogRecord to ReadableLogRecord before passing to BatchProcessor
        # Note: resource should not be None at this point as it's set during Logger.emit()
        resource = (
            log_record.resource
            if log_record.resource is not None
            else Resource.create({})
        )
        readable_log_record = ReadableLogRecord(
            log_record=log_record.log_record,
            resource=resource,
            instrumentation_scope=log_record.instrumentation_scope,
            limits=log_record.limits,
        )
        return self._batch_processor.emit(readable_log_record)

    def shutdown(self):
        return self._batch_processor.shutdown()

    def force_flush(self, timeout_millis: int | None = None) -> bool:
        return self._batch_processor.force_flush(timeout_millis)

    @staticmethod
    def _default_max_queue_size():
        try:
            return int(
                environ.get(OTEL_BLRP_MAX_QUEUE_SIZE, _DEFAULT_MAX_QUEUE_SIZE)
            )
        except ValueError:
            _logger.exception(
                _ENV_VAR_INT_VALUE_ERROR_MESSAGE,
                OTEL_BLRP_MAX_QUEUE_SIZE,
                _DEFAULT_MAX_QUEUE_SIZE,
            )
            return _DEFAULT_MAX_QUEUE_SIZE

    @staticmethod
    def _default_schedule_delay_millis():
        try:
            return int(
                environ.get(
                    OTEL_BLRP_SCHEDULE_DELAY, _DEFAULT_SCHEDULE_DELAY_MILLIS
                )
            )
        except ValueError:
            _logger.exception(
                _ENV_VAR_INT_VALUE_ERROR_MESSAGE,
                OTEL_BLRP_SCHEDULE_DELAY,
                _DEFAULT_SCHEDULE_DELAY_MILLIS,
            )
            return _DEFAULT_SCHEDULE_DELAY_MILLIS

    @staticmethod
    def _default_max_export_batch_size():
        try:
            return int(
                environ.get(
                    OTEL_BLRP_MAX_EXPORT_BATCH_SIZE,
                    _DEFAULT_MAX_EXPORT_BATCH_SIZE,
                )
            )
        except ValueError:
            _logger.exception(
                _ENV_VAR_INT_VALUE_ERROR_MESSAGE,
                OTEL_BLRP_MAX_EXPORT_BATCH_SIZE,
                _DEFAULT_MAX_EXPORT_BATCH_SIZE,
            )
            return _DEFAULT_MAX_EXPORT_BATCH_SIZE

    @staticmethod
    def _default_export_timeout_millis():
        try:
            return int(
                environ.get(
                    OTEL_BLRP_EXPORT_TIMEOUT, _DEFAULT_EXPORT_TIMEOUT_MILLIS
                )
            )
        except ValueError:
            _logger.exception(
                _ENV_VAR_INT_VALUE_ERROR_MESSAGE,
                OTEL_BLRP_EXPORT_TIMEOUT,
                _DEFAULT_EXPORT_TIMEOUT_MILLIS,
            )
            return _DEFAULT_EXPORT_TIMEOUT_MILLIS

    @staticmethod
    def _validate_arguments(
        max_queue_size, schedule_delay_millis, max_export_batch_size
    ):
        if max_queue_size <= 0:
            raise ValueError("max_queue_size must be a positive integer.")

        if schedule_delay_millis <= 0:
            raise ValueError("schedule_delay_millis must be positive.")

        if max_export_batch_size <= 0:
            raise ValueError(
                "max_export_batch_size must be a positive integer."
            )

        if max_export_batch_size > max_queue_size:
            raise ValueError(
                "max_export_batch_size must be less than or equal to max_queue_size."
            )



"""
Package support for the REST example
"""

import logging

import salt.utils.data
import salt.utils.platform

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pkg"


def __virtual__():
    """
    Only work on systems that are a proxy minion
    """
    try:
        if (
            salt.utils.platform.is_proxy()
            and __opts__["proxy"]["proxytype"] == "rest_sample"
        ):
            return __virtualname__
    except KeyError:
        return (
            False,
            "The rest_package execution module failed to load. Check the "
            "proxy key in pillar.",
        )

    return (
        False,
        "The rest_package execution module failed to load: only works on a "
        "rest_sample proxy minion.",
    )


def list_pkgs(versions_as_list=False, **kwargs):
    return __proxy__["rest_sample.package_list"]()


def install(name=None, refresh=False, fromrepo=None, pkgs=None, sources=None, **kwargs):
    return __proxy__["rest_sample.package_install"](name, **kwargs)


def remove(name=None, pkgs=None, **kwargs):
    return __proxy__["rest_sample.package_remove"](name)


def version(*names, **kwargs):
    """
    Returns a string representing the package version or an empty string if not
    installed. If more than one package name is specified, a dict of
    name/version pairs is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version <package name>
        salt '*' pkg.version <package1> <package2> <package3> ...
    """
    if len(names) == 1:
        return str(__proxy__["rest_sample.package_status"](names[0]))


def upgrade(refresh=True, skip_verify=True, **kwargs):
    old = __proxy__["rest_sample.package_list"]()
    new = __proxy__["rest_sample.uptodate"]()
    pkg_installed = __proxy__["rest_sample.upgrade"]()
    return salt.utils.data.compare_dicts(old, pkg_installed)


def installed(
    name,
    version=None,
    refresh=False,
    fromrepo=None,
    skip_verify=False,
    pkgs=None,
    sources=None,
    **kwargs,
):

    p = __proxy__["rest_sample.package_status"](name)
    if version is None:
        if "ret" in p:
            return str(p["ret"])
        else:
            return True
    else:
        if p is not None:
            return version == str(p)



"""
Installation of Python Packages Using pip
=========================================

These states manage system installed python packages. Note that pip must be
installed for these states to be available, so pip states should include a
requisite to a pkg.installed state for the package which provides pip
(``python-pip`` in most cases). Example:

.. code-block:: yaml

    python-pip:
      pkg.installed

    virtualenvwrapper:
      pip.installed:
        - require:
          - pkg: python-pip
"""

import logging
import sys
import types

import salt.utils.data
import salt.utils.versions
from salt.exceptions import CommandExecutionError, CommandNotFoundError


def purge_pip():
    """
    Purge pip and its sub-modules
    """
    # Remove references to the loaded pip module above so reloading works
    if "pip" not in sys.modules:
        return
    pip_related_entries = [
        (k, v)
        for (k, v) in sys.modules.items()
        if getattr(v, "__module__", "").startswith("pip.")
        or (isinstance(v, types.ModuleType) and v.__name__.startswith("pip."))
    ]
    for name, entry in pip_related_entries:
        sys.modules.pop(name)
        del entry

    if "pip" in globals():
        del globals()["pip"]
    if "pip" in locals():
        del locals()["pip"]
    sys_modules_pip = sys.modules.pop("pip", None)
    if sys_modules_pip is not None:
        del sys_modules_pip


def pip_has_internal_exceptions_mod(ver):
    """
    True when the pip version has the `pip._internal.exceptions` module
    """
    return salt.utils.versions.compare(
        ver1=ver,
        oper=">=",
        ver2="10.0",
    )


def pip_has_exceptions_mod(ver):
    """
    True when the pip version has the `pip.exceptions` module
    """
    if pip_has_internal_exceptions_mod(ver):
        return False
    return salt.utils.versions.compare(ver1=ver, oper=">=", ver2="1.0")


try:
    import pip

    HAS_PIP = True
except ImportError:
    HAS_PIP = False
    purge_pip()


if HAS_PIP is True:
    if not hasattr(purge_pip, "__pip_ver__"):
        purge_pip.__pip_ver__ = pip.__version__
    elif purge_pip.__pip_ver__ != pip.__version__:
        purge_pip()
        import pip

        purge_pip.__pip_ver__ = pip.__version__
    if salt.utils.versions.compare(ver1=pip.__version__, oper=">=", ver2="10.0"):
        from pip._internal.exceptions import (  # pylint: disable=E0611,E0401
            InstallationError,
        )
    elif salt.utils.versions.compare(ver1=pip.__version__, oper=">=", ver2="1.0"):
        from pip.exceptions import InstallationError  # pylint: disable=E0611,E0401
    else:
        InstallationError = ValueError


# pylint: enable=import-error

logger = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pip"


def _from_line(*args, **kwargs):
    import pip

    if salt.utils.versions.compare(ver1=pip.__version__, oper=">=", ver2="18.1"):
        import pip._internal.req.constructors  # pylint: disable=E0611,E0401

        return pip._internal.req.constructors.install_req_from_line(*args, **kwargs)
    elif salt.utils.versions.compare(ver1=pip.__version__, oper=">=", ver2="10.0"):
        import pip._internal.req  # pylint: disable=E0611,E0401

        return pip._internal.req.InstallRequirement.from_line(*args, **kwargs)
    else:
        import pip.req  # pylint: disable=E0611,E0401

        return pip.req.InstallRequirement.from_line(*args, **kwargs)


def __virtual__():
    """
    Only load if the pip module is available in __salt__
    """
    if "pip.list" in __salt__:
        return __virtualname__
    return False


def _fulfills_version_spec(version, version_spec):
    """
    Check version number against version specification info and return a
    boolean value based on whether or not the version number meets the
    specified version.
    """
    try:
        from packaging.specifiers import InvalidSpecifier, SpecifierSet
        from packaging.version import InvalidVersion

        # Build a SpecifierSet string from the version_spec list of tuples
        specs = []
        for oper, spec in version_spec:
            if oper is not None:
                specs.append(f"{oper}{spec}")

        if not specs:
            return True

        spec_set = SpecifierSet(",".join(specs))
        return spec_set.contains(version)
    except (ImportError, InvalidVersion, InvalidSpecifier):
        # Fallback to the old logic if packaging is not available
        # or if the version/spec is not PEP 440 compliant
        for oper, spec in version_spec:
            if oper is None:
                continue
            if not salt.utils.versions.compare(
                ver1=version, oper=oper, ver2=spec, cmp_func=_pep440_version_cmp
            ):
                return False
        return True


def _check_pkg_version_format(pkg):
    """
    Takes a package name and version specification (if any) and checks it using
    the pip library.
    """

    ret = {"result": False, "comment": None, "prefix": None, "version_spec": None}

    if not HAS_PIP:
        ret["comment"] = (
            "An importable Python pip module is required but could not be "
            "found on your system. This usually means that the system's pip "
            "package is not installed properly."
        )

        return ret

    from_vcs = False
    try:
        # Get the requirement object from the pip library
        try:
            # With pip < 1.2, the __version__ attribute does not exist and
            # vcs+URL urls are not properly parsed.
            # The next line is meant to trigger an AttributeError and
            # handle lower pip versions
            logger.debug("Installed pip version: %s", pip.__version__)
            install_req = _from_line(pkg)
        except AttributeError:
            logger.debug("Installed pip version is lower than 1.2")
            supported_vcs = ("git", "svn", "hg", "bzr")
            if pkg.startswith(supported_vcs):
                for vcs in supported_vcs:
                    if pkg.startswith(vcs):
                        from_vcs = True
                        install_req = _from_line(pkg.split(f"{vcs}+")[-1])
                        break
            else:
                install_req = _from_line(pkg)
    except (ValueError, InstallationError) as exc:
        ret["result"] = False
        if not from_vcs and "=" in pkg and "==" not in pkg:
            ret["comment"] = (
                "Invalid version specification in package {}. '=' is "
                "not supported, use '==' instead.".format(pkg)
            )
            return ret
        ret["comment"] = "pip raised an exception while parsing '{}': {}".format(
            pkg, exc
        )
        return ret

    if install_req.req is None:
        # This is most likely an url and there's no way to know what will
        # be installed before actually installing it.
        ret["result"] = True
        ret["prefix"] = ""
        ret["version_spec"] = []
    else:
        ret["result"] = True
        normalize = __salt__["pip.normalize"]
        try:
            ret["prefix"] = normalize(install_req.req.project_name)
            ret["version_spec"] = install_req.req.specs
        except Exception:  # pylint: disable=broad-except
            ret["prefix"] = normalize(install_req.name)
            if hasattr(install_req, "specifier"):
                specifier = install_req.specifier
            else:
                specifier = install_req.req.specifier
            ret["version_spec"] = [(spec.operator, spec.version) for spec in specifier]

    return ret


def _check_if_installed(
    prefix,
    state_pkg_name,
    version_spec,
    ignore_installed,
    force_reinstall,
    upgrade,
    user,
    cwd,
    bin_env,
    env_vars,
    index_url,
    extra_index_url,
    pip_list=False,
    **kwargs,
):
    """
    Takes a package name and version specification (if any) and checks it is
    installed

    Keyword arguments include:
        pip_list: optional dict of installed pip packages, and their versions,
            to search through to check if the package is installed. If not
            provided, one will be generated in this function by querying the
            system.

    Returns:
     result: None means the command failed to run
     result: True means the package is installed
     result: False means the package is not installed
    """
    ret = {"result": False, "comment": None}

    # If we are not passed a pip list, get one:
    pip_list = salt.utils.data.CaseInsensitiveDict(
        pip_list
        or __salt__["pip.list"](
            prefix, bin_env=bin_env, user=user, cwd=cwd, env_vars=env_vars, **kwargs
        )
    )

    # If the package was already installed, check
    # the ignore_installed and force_reinstall flags
    if ignore_installed is False and prefix in pip_list:
        if force_reinstall is False and not upgrade:
            # Check desired version (if any) against currently-installed
            if (
                any(version_spec)
                and _fulfills_version_spec(pip_list[prefix], version_spec)
            ) or (not any(version_spec)):
                ret["result"] = True
                ret["comment"] = "Python package {} was already installed".format(
                    state_pkg_name
                )
                return ret
        if force_reinstall is False and upgrade:
            # Check desired version (if any) against currently-installed
            include_alpha = False
            include_beta = False
            include_rc = False
            if any(version_spec):
                for spec in version_spec:
                    if "a" in spec[1]:
                        include_alpha = True
                    if "b" in spec[1]:
                        include_beta = True
                    if "rc" in spec[1]:
                        include_rc = True
            available_versions = __salt__["pip.list_all_versions"](
                prefix,
                bin_env=bin_env,
                include_alpha=include_alpha,
                include_beta=include_beta,
                include_rc=include_rc,
                user=user,
                cwd=cwd,
                index_url=index_url,
                extra_index_url=extra_index_url,
            )
            desired_version = ""
            if any(version_spec) and available_versions:
                for version in reversed(available_versions):
                    if _fulfills_version_spec(version, version_spec):
                        desired_version = version
                        break
            elif available_versions:
                desired_version = available_versions[-1]
            if not desired_version:
                ret["result"] = True
                ret["comment"] = (
                    "Python package {} was already "
                    "installed and\nthe available upgrade "
                    "doesn't fulfills the version "
                    "requirements".format(prefix)
                )
                return ret
            if _pep440_version_cmp(pip_list[prefix], desired_version) == 0:
                ret["result"] = True
                ret["comment"] = "Python package {} was already installed".format(
                    state_pkg_name
                )
                return ret

    return ret


def _pep440_version_cmp(pkg1, pkg2, ignore_epoch=False):
    """
    Compares two version strings using packaging.version.Version.
    Return -1 if version1 < version2, 0 if version1 ==version2,
    and 1 if version1 > version2. Return None if there was a problem
    making the comparison.
    """

    def normalize(x):
        return str(x).split("!", 1)[-1] if ignore_epoch else str(x)

    pkg1 = normalize(pkg1)
    pkg2 = normalize(pkg2)

    try:
        if salt.utils.versions.Version(pkg1) < salt.utils.versions.Version(pkg2):
            return -1
        if salt.utils.versions.Version(pkg1) == salt.utils.versions.Version(pkg2):
            return 0
        if salt.utils.versions.Version(pkg1) > salt.utils.versions.Version(pkg2):
            return 1
    except Exception as exc:  # pylint: disable=broad-except
        logger.debug(
            'Comparison of package versions "%s" and "%s" failed: %s', pkg1, pkg2, exc
        )
    return None


def installed(
    name,
    pkgs=None,
    pip_bin=None,
    requirements=None,
    bin_env=None,
    use_wheel=False,
    no_use_wheel=False,
    log=None,
    proxy=None,
    timeout=None,
    repo=None,
    editable=None,
    find_links=None,
    index_url=None,
    extra_index_url=None,
    no_index=False,
    mirrors=None,
    build=None,
    target=None,
    download=None,
    download_cache=None,
    source=None,
    upgrade=False,
    force_reinstall=False,
    ignore_installed=False,
    exists_action=None,
    no_deps=False,
    no_install=False,
    no_download=False,
    install_options=None,
    global_options=None,
    user=None,
    cwd=None,
    pre_releases=False,
    cert=None,
    allow_all_external=False,
    allow_external=None,
    allow_unverified=None,
    process_dependency_links=False,
    env_vars=None,
    use_vt=False,
    trusted_host=None,
    no_cache_dir=False,
    cache_dir=None,
    no_binary=None,
    extra_args=None,
    **kwargs,
):
    """
    Make sure the package is installed

    name
        The name of the python package to install. You can also specify version
        numbers here using the standard operators ``==, >=, <=``. If
        ``requirements`` or ``pkgs`` is given, this parameter will be ignored.

        Example:

        .. code-block:: yaml

            django:
              pip.installed:
                - name: django >= 1.6, <= 1.7
                - require:
                  - pkg: python-pip

        Installs the latest Django version greater than 1.6 but less
        than 1.7.

    pkgs
        A list of python packages to install. This let you install multiple
        packages at the same time.

        Example:

        .. code-block:: yaml

            django-and-psycopg2:
              pip.installed:
                - pkgs:
                  - django >= 1.6, <= 1.7
                  - psycopg2 >= 2.8.4
                - require:
                  - pkg: python-pip

        Installs the latest Django version greater than 1.6 but less than 1.7
        and the latest psycopg2 greater than 2.8.4 at the same time.

    requirements
        Path to a pip requirements file. If the path begins with salt://
        the file will be transferred from the master file server.

    user
        The user under which to run pip

    use_wheel : False
        Prefer wheel archives (requires pip>=1.4)

    no_use_wheel : False
        Force to not use wheel archives (requires pip>=1.4)

    no_binary
        Force to not use binary packages (requires pip >= 7.0.0)
        Accepts either :all: to disable all binary packages, :none: to empty the set,
        or a list of one or more packages

    Example:

    .. code-block:: yaml

        django:
          pip.installed:
            - no_binary: ':all:'

        flask:
          pip.installed:
            - no_binary:
              - itsdangerous
              - click

    log
        Log file where a complete (maximum verbosity) record will be kept

    proxy
        Specify a proxy in the form
        user:passwd@proxy.server:port. Note that the
        user:password@ is optional and required only if you
        are behind an authenticated proxy.  If you provide
        user@proxy.server:port then you will be prompted for a
        password.

    timeout
        Set the socket timeout (default 15 seconds)

    editable
        install something editable (i.e.
        git+https://github.com/worldcompany/djangoembed.git#egg=djangoembed)

    find_links
        URL to look for packages at

    index_url
        Base URL of Python Package Index

    extra_index_url
        Extra URLs of package indexes to use in addition to ``index_url``

    no_index
        Ignore package index

    mirrors
        Specific mirror URL(s) to query (automatically adds --use-mirrors)

    build
        Unpack packages into ``build`` dir

    target
        Install packages into ``target`` dir

    download
        Download packages into ``download`` instead of installing them

    download_cache
        Cache downloaded packages in ``download_cache`` dir

    source
        Check out ``editable`` packages into ``source`` dir

    upgrade
        Upgrade all packages to the newest available version

    force_reinstall
        When upgrading, reinstall all packages even if they are already
        up-to-date.

    ignore_installed
        Ignore the installed packages (reinstalling instead)

    exists_action
        Default action when a path already exists: (s)witch, (i)gnore, (w)ipe,
        (b)ackup

    no_deps
        Ignore package dependencies

    no_install
        Download and unpack all packages, but don't actually install them

    no_cache_dir:
        Disable the cache.

    cwd
        Current working directory to run pip from

    pre_releases
        Include pre-releases in the available versions

    cert
        Provide a path to an alternate CA bundle

    allow_all_external
        Allow the installation of all externally hosted files

    allow_external
        Allow the installation of externally hosted files (comma separated list)

    allow_unverified
        Allow the installation of insecure and unverifiable files (comma separated list)

    process_dependency_links
        Enable the processing of dependency links

    env_vars
        Add or modify environment variables. Useful for tweaking build steps,
        such as specifying INCLUDE or LIBRARY paths in Makefiles, build scripts or
        compiler calls.  This must be in the form of a dictionary or a mapping.

        Example:

        .. code-block:: yaml

            django:
              pip.installed:
                - name: django_app
                - env_vars:
                    CUSTOM_PATH: /opt/django_app
                    VERBOSE: True

    use_vt
        Use VT terminal emulation (see output while installing)

    trusted_host
        Mark this host as trusted, even though it does not have valid or any
        HTTPS.

    bin_env : None
        Absolute path to a virtual environment directory or absolute path to
        a pip executable. The example below assumes a virtual environment
        has been created at ``/foo/.virtualenvs/bar``.

        Example:

        .. code-block:: yaml

            django:
            pip.installed:
                - name: django >= 1.6, <= 1.7
                - bin_env: /foo/.virtualenvs/bar
                - require:
                - pkg: python-pip

        Or

        Example:

        .. code-block:: yaml

            django:
            pip.installed:
                - name: django >= 1.6, <= 1.7
                - bin_env: /foo/.virtualenvs/bar/bin/pip
                - require:
                - pkg: python-pip

    .. admonition:: Attention

        The following arguments are deprecated, do not use.

    pip_bin : None
        Deprecated, use ``bin_env``

    .. versionchanged:: 0.17.0
        ``use_wheel`` option added.

    install_options

        Extra arguments to be supplied to the setup.py install command.
        If you are using an option with a directory path, be sure to use
        absolute path.

        Example:

        .. code-block:: yaml

            django:
              pip.installed:
                - name: django
                - install_options:
                  - --prefix=/blah
                - require:
                  - pkg: python-pip

    global_options
        Extra global options to be supplied to the setup.py call before the
        install command.

        .. versionadded:: 2014.1.3

    .. admonition:: Attention

        As of Salt 0.17.0 the pip state **needs** an importable pip module.
        This usually means having the system's pip package installed or running
        Salt from an active `virtualenv`_.

        The reason for this requirement is because ``pip`` already does a
        pretty good job parsing its own requirements. It makes no sense for
        Salt to do ``pip`` requirements parsing and validation before passing
        them to the ``pip`` library. It's functionality duplication and it's
        more error prone.


    .. admonition:: Attention

        Please set ``reload_modules: True`` to have the salt minion
        import this module after installation.


    Example:

    .. code-block:: yaml

        pyopenssl:
            pip.installed:
                - name: pyOpenSSL
                - reload_modules: True
                - exists_action: i

    extra_args
        pip keyword and positional arguments not yet implemented in salt

        .. code-block:: yaml

            pandas:
              pip.installed:
                - name: pandas
                - extra_args:
                  - --latest-pip-kwarg: param
                  - --latest-pip-arg

        .. warning::

            If unsupported options are passed here that are not supported in a
            minion's version of pip, a `No such option error` will be thrown.


    .. _`virtualenv`: http://www.virtualenv.org/en/latest/

    If you are using onedir packages and you need to install python packages into
    the system python environment, you must provide the pip_bin or
    bin_env to the pip state module.


    .. code-block:: yaml

        lib-foo:
          pip.installed:
            - pip_bin: /usr/bin/pip3
        lib-bar:
          pip.installed:
            - bin_env: /usr/bin/python3
    """
    if pip_bin and not bin_env:
        bin_env = pip_bin

    # If pkgs is present, ignore name
    if pkgs:
        if not isinstance(pkgs, list):
            return {
                "name": name,
                "result": False,
                "changes": {},
                "comment": "pkgs argument must be formatted as a list",
            }
    else:
        pkgs = [name]

    # Assumption: If `pkg` is not an `string`, it's a `collections.OrderedDict`
    # prepro = lambda pkg: pkg if type(pkg) == str else \
    #     ' '.join((pkg.items()[0][0], pkg.items()[0][1].replace(',', ';')))
    # pkgs = ','.join([prepro(pkg) for pkg in pkgs])
    def prepro(pkg):
        return (
            pkg
            if isinstance(pkg, str)
            else " ".join((pkg.items()[0][0], pkg.items()[0][1]))
        )

    pkgs = [prepro(pkg) for pkg in pkgs]

    ret = {"name": ";".join(pkgs), "result": None, "comment": "", "changes": {}}

    try:
        cur_version = __salt__["pip.version"](bin_env)
    except (CommandNotFoundError, CommandExecutionError) as err:
        ret["result"] = False
        ret["comment"] = f"Error installing '{name}': {err}"
        return ret
    # Check that the pip binary supports the 'use_wheel' option
    if use_wheel:
        min_version = "1.4"
        max_version = "9.0.3"
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        too_high = salt.utils.versions.compare(
            ver1=cur_version, oper=">", ver2=max_version
        )
        if too_low or too_high:
            ret["result"] = False
            ret["comment"] = (
                "The 'use_wheel' option is only supported in "
                "pip between {} and {}. The version of pip detected "
                "was {}.".format(min_version, max_version, cur_version)
            )
            return ret

    # Check that the pip binary supports the 'no_use_wheel' option
    if no_use_wheel:
        min_version = "1.4"
        max_version = "9.0.3"
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        too_high = salt.utils.versions.compare(
            ver1=cur_version, oper=">", ver2=max_version
        )
        if too_low or too_high:
            ret["result"] = False
            ret["comment"] = (
                "The 'no_use_wheel' option is only supported in "
                "pip between {} and {}. The version of pip detected "
                "was {}.".format(min_version, max_version, cur_version)
            )
            return ret

    # Check that the pip binary supports the 'no_binary' option
    if no_binary:
        min_version = "7.0.0"
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        if too_low:
            ret["result"] = False
            ret["comment"] = (
                "The 'no_binary' option is only supported in "
                "pip {} and newer. The version of pip detected "
                "was {}.".format(min_version, cur_version)
            )
            return ret

    # Get the packages parsed name and version from the pip library.
    # This only is done when there is no requirements or editable parameter.
    pkgs_details = []
    if pkgs and not (requirements or editable):
        comments = []
        for pkg in iter(pkgs):
            out = _check_pkg_version_format(pkg)
            if out["result"] is False:
                ret["result"] = False
                comments.append(out["comment"])
            elif out["result"] is True:
                pkgs_details.append((out["prefix"], pkg, out["version_spec"]))

        if ret["result"] is False:
            ret["comment"] = "\n".join(comments)
            return ret

    # If the user does not exist, stop here with error:
    if user and "user.info" in __salt__ and not __salt__["user.info"](user):
        # The user does not exists, exit with result set to False
        ret["result"] = False
        ret["comment"] = f"User {user} does not exist"
        return ret

    # If a requirements file is specified, only install the contents of the
    # requirements file. Similarly, using the --editable flag with pip should
    # also ignore the "name" and "pkgs" parameters.
    target_pkgs = []
    already_installed_comments = []
    if requirements or editable:
        comments = []
        # Append comments if this is a dry run.
        if __opts__["test"]:
            ret["result"] = None
            if requirements:
                # TODO: Check requirements file against currently-installed
                # packages to provide more accurate state output.
                comments.append(
                    f"Requirements file '{requirements}' will be processed."
                )
            if editable:
                comments.append(
                    "Package will be installed in editable mode (i.e. "
                    'setuptools "develop mode") from {}.'.format(editable)
                )
            ret["comment"] = " ".join(comments)
            return ret

    # No requirements case.
    # Check pre-existence of the requested packages.
    else:
        # Attempt to pre-cache the current pip list
        try:
            pip_list = __salt__["pip.list"](
                bin_env=bin_env, user=user, cwd=cwd, env_vars=env_vars
            )
        # If we fail, then just send False, and we'll try again in the next function call
        except Exception as exc:  # pylint: disable=broad-except
            logger.warning(
                "Pre-caching of PIP packages during states.pip.installed failed by exception from pip.list: %s",
                exc,
                exc_info=True,
            )
            pip_list = False

        for prefix, state_pkg_name, version_spec in pkgs_details:
            if prefix:
                out = _check_if_installed(
                    prefix,
                    state_pkg_name,
                    version_spec,
                    ignore_installed,
                    force_reinstall,
                    upgrade,
                    user,
                    cwd,
                    bin_env,
                    env_vars,
                    index_url,
                    extra_index_url,
                    pip_list,
                    **kwargs,
                )
                # If _check_if_installed result is None, something went wrong with
                # the command running. This way we keep stateful output.
                if out["result"] is None:
                    ret["result"] = False
                    ret["comment"] = out["comment"]
                    return ret
            else:
                out = {"result": False, "comment": None}

            result = out["result"]

            # The package is not present. Add it to the pkgs to install.
            if result is False:
                # Replace commas (used for version ranges) with semicolons
                # (which are not supported) in name so it does not treat
                # them as multiple packages.
                target_pkgs.append((prefix, state_pkg_name.replace(",", ";")))

                # Append comments if this is a dry run.
                if __opts__["test"]:
                    # If there is more than one package, compute
                    # the total amount and exit
                    if len(pkgs_details) > 1:
                        msg = "Python package(s) set to be installed:"
                        for pkg in pkgs_details:
                            msg += "\n"
                            msg += pkg[1]
                            ret["comment"] = msg
                    else:
                        msg = "Python package {0} is set to be installed"
                        ret["comment"] = msg.format(state_pkg_name)
                    ret["result"] = None
                    return ret

            # The package is already present and will not be reinstalled.
            elif result is True:
                # Append comment stating its presence
                already_installed_comments.append(out["comment"])

            # The command pip.list failed. Abort.
            elif result is None:
                ret["result"] = None
                ret["comment"] = out["comment"]
                return ret

        # No packages to install.
        if not target_pkgs:
            ret["result"] = True
            aicomms = "\n".join(already_installed_comments)
            last_line = "All specified packages are already installed" + (
                " and up-to-date" if upgrade else ""
            )
            ret["comment"] = aicomms + ("\n" if aicomms else "") + last_line
            return ret

    # Construct the string that will get passed to the install call
    pkgs_str = ",".join([state_name for _, state_name in target_pkgs])

    # Call to install the package. Actual installation takes place here
    pip_install_call = __salt__["pip.install"](
        pkgs=f"{pkgs_str}" if pkgs_str else "",
        requirements=requirements,
        bin_env=bin_env,
        use_wheel=use_wheel,
        no_use_wheel=no_use_wheel,
        no_binary=no_binary,
        log=log,
        proxy=proxy,
        timeout=timeout,
        editable=editable,
        find_links=find_links,
        index_url=index_url,
        extra_index_url=extra_index_url,
        no_index=no_index,
        mirrors=mirrors,
        build=build,
        target=target,
        download=download,
        download_cache=download_cache,
        source=source,
        upgrade=upgrade,
        force_reinstall=force_reinstall,
        ignore_installed=ignore_installed,
        exists_action=exists_action,
        no_deps=no_deps,
        no_install=no_install,
        no_download=no_download,
        install_options=install_options,
        global_options=global_options,
        user=user,
        cwd=cwd,
        pre_releases=pre_releases,
        cert=cert,
        allow_all_external=allow_all_external,
        allow_external=allow_external,
        allow_unverified=allow_unverified,
        process_dependency_links=process_dependency_links,
        saltenv=__env__,
        env_vars=env_vars,
        use_vt=use_vt,
        trusted_host=trusted_host,
        no_cache_dir=no_cache_dir,
        extra_args=extra_args,
        disable_version_check=True,
        **kwargs,
    )

    if pip_install_call and pip_install_call.get("retcode", 1) == 0:
        ret["result"] = True

        if requirements or editable:
            comments = []
            if requirements:
                PIP_REQUIREMENTS_NOCHANGE = [
                    "Requirement already satisfied",
                    "Requirement already up-to-date",
                    "Requirement not upgraded",
                    "Collecting",
                    "Cloning",
                    "Cleaning up...",
                    "Looking in indexes",
                ]
                for line in pip_install_call.get("stdout", "").split("\n"):
                    if not any(
                        [line.strip().startswith(x) for x in PIP_REQUIREMENTS_NOCHANGE]
                    ):
                        ret["changes"]["requirements"] = True
                if ret["changes"].get("requirements"):
                    comments.append(
                        "Successfully processed requirements file {}.".format(
                            requirements
                        )
                    )
                else:
                    comments.append("Requirements were already installed.")

            if editable:
                comments.append(
                    "Package successfully installed from VCS checkout {}.".format(
                        editable
                    )
                )
                ret["changes"]["editable"] = True
            ret["comment"] = " ".join(comments)
        else:
            # Check that the packages set to be installed were installed.
            # Create comments reporting success and failures
            pkg_404_comms = []

            already_installed_packages = set()
            for line in pip_install_call.get("stdout", "").split("\n"):
                # Output for already installed packages:
                # 'Requirement already up-to-date: jinja2 in /usr/local/lib/python2.7/dist-packages\nCleaning up...'
                if line.startswith("Requirement already up-to-date: "):
                    package = line.split(":", 1)[1].split()[0]
                    already_installed_packages.add(package.lower())

            for prefix, state_name in target_pkgs:
                # Case for packages that are not an URL
                if prefix:
                    pipsearch = salt.utils.data.CaseInsensitiveDict(
                        __salt__["pip.list"](
                            prefix,
                            bin_env,
                            user=user,
                            cwd=cwd,
                            env_vars=env_vars,
                            **kwargs,
                        )
                    )

                    # If we didn't find the package in the system after
                    # installing it report it
                    if not pipsearch:
                        pkg_404_comms.append(
                            "There was no error installing package '{}' "
                            "although it does not show when calling "
                            "'pip.freeze'.".format(pkg)
                        )
                    else:
                        if (
                            prefix in pipsearch
                            and prefix.lower() not in already_installed_packages
                        ):
                            ver = pipsearch[prefix]
                            ret["changes"][f"{prefix}=={ver}"] = "Installed"
                # Case for packages that are an URL
                else:
                    ret["changes"][f"{state_name}==???"] = "Installed"

            # Set comments
            aicomms = "\n".join(already_installed_comments)
            succ_comm = (
                "All packages were successfully installed"
                if not pkg_404_comms
                else "\n".join(pkg_404_comms)
            )
            ret["comment"] = aicomms + ("\n" if aicomms else "") + succ_comm

            return ret

    elif pip_install_call:
        ret["result"] = False
        if "stdout" in pip_install_call:
            error = "Error: {} {}".format(
                pip_install_call["stdout"], pip_install_call["stderr"]
            )
        else:
            error = "Error: {}".format(pip_install_call["comment"])

        if requirements or editable:
            comments = []
            if requirements:
                comments.append(f'Unable to process requirements file "{requirements}"')
            if editable:
                comments.append(f"Unable to install from VCS checkout {editable}.")
            comments.append(error)
            ret["comment"] = " ".join(comments)
        else:
            pkgs_str = ", ".join([state_name for _, state_name in target_pkgs])
            aicomms = "\n".join(already_installed_comments)
            error_comm = f"Failed to install packages: {pkgs_str}. {error}"
            ret["comment"] = aicomms + ("\n" if aicomms else "") + error_comm
    else:
        ret["result"] = False
        ret["comment"] = "Could not install package"

    return ret


def removed(
    name,
    requirements=None,
    bin_env=None,
    log=None,
    proxy=None,
    timeout=None,
    user=None,
    cwd=None,
    use_vt=False,
):
    """
    Make sure that a package is not installed.

    name
        The name of the package to uninstall
    user
        The user under which to run pip
    bin_env : None
        the pip executable or virtualenenv to use
    use_vt
        Use VT terminal emulation (see output while installing)
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    try:
        pip_list = __salt__["pip.list"](bin_env=bin_env, user=user, cwd=cwd)
    except (CommandExecutionError, CommandNotFoundError) as err:
        ret["result"] = False
        ret["comment"] = f"Error uninstalling '{name}': {err}"
        return ret

    if name not in pip_list:
        ret["result"] = True
        ret["comment"] = "Package is not installed."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Package {name} is set to be removed"
        return ret

    if __salt__["pip.uninstall"](
        pkgs=name,
        requirements=requirements,
        bin_env=bin_env,
        log=log,
        proxy=proxy,
        timeout=timeout,
        user=user,
        cwd=cwd,
        use_vt=use_vt,
    ):
        ret["result"] = True
        ret["changes"][name] = "Removed"
        ret["comment"] = "Package was successfully removed."
    else:
        ret["result"] = False
        ret["comment"] = "Could not remove package."
    return ret


def uptodate(name, bin_env=None, user=None, cwd=None, use_vt=False):
    """
    .. versionadded:: 2015.5.0

    Verify that the system is completely up to date.

    name
        The name has no functional value and is only used as a tracking
        reference
    user
        The user under which to run pip
    bin_env
        the pip executable or virtualenenv to use
    use_vt
        Use VT terminal emulation (see output while installing)
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": "Failed to update."}

    try:
        packages = __salt__["pip.list_upgrades"](bin_env=bin_env, user=user, cwd=cwd)
    except Exception as e:  # pylint: disable=broad-except
        ret["comment"] = str(e)
        return ret

    if not packages:
        ret["comment"] = "System is already up-to-date."
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = "System update will be performed"
        ret["result"] = None
        return ret

    updated = __salt__["pip.upgrade"](
        bin_env=bin_env, user=user, cwd=cwd, use_vt=use_vt
    )

    if updated.get("result") is False:
        ret.update(updated)
    elif updated:
        ret["changes"] = updated
        ret["comment"] = "Upgrade successful."
        ret["result"] = True
    else:
        ret["comment"] = "Upgrade failed."

    return ret



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing_extensions import deprecated


@deprecated(
    "Use metrics defined in the :py:const:`opentelemetry.semconv.metrics` and :py:const:`opentelemetry.semconv._incubating.metrics` modules instead. Deprecated since version 1.25.0.",
)
class MetricInstruments:
    SCHEMA_URL = "https://opentelemetry.io/schemas/1.21.0"
    """
    The URL of the OpenTelemetry schema for these keys and values.
    """

    HTTP_SERVER_DURATION = "http.server.duration"
    """
    Measures the duration of inbound HTTP requests
    Instrument: histogram
    Unit: s
    """

    HTTP_SERVER_ACTIVE_REQUESTS = "http.server.active_requests"
    """
    Measures the number of concurrent HTTP requests that are currently in-flight
    Instrument: updowncounter
    Unit: {request}
    """

    HTTP_SERVER_REQUEST_SIZE = "http.server.request.size"
    """
    Measures the size of HTTP request messages (compressed)
    Instrument: histogram
    Unit: By
    """

    HTTP_SERVER_RESPONSE_SIZE = "http.server.response.size"
    """
    Measures the size of HTTP response messages (compressed)
    Instrument: histogram
    Unit: By
    """

    HTTP_CLIENT_DURATION = "http.client.duration"
    """
    Measures the duration of outbound HTTP requests
    Instrument: histogram
    Unit: s
    """

    HTTP_CLIENT_REQUEST_SIZE = "http.client.request.size"
    """
    Measures the size of HTTP request messages (compressed)
    Instrument: histogram
    Unit: By
    """

    HTTP_CLIENT_RESPONSE_SIZE = "http.client.response.size"
    """
    Measures the size of HTTP response messages (compressed)
    Instrument: histogram
    Unit: By
    """

    PROCESS_RUNTIME_JVM_MEMORY_INIT = "process.runtime.jvm.memory.init"
    """
    Measure of initial memory requested
    Instrument: updowncounter
    Unit: By
    """

    PROCESS_RUNTIME_JVM_SYSTEM_CPU_UTILIZATION = (
        "process.runtime.jvm.system.cpu.utilization"
    )
    """
    Recent CPU utilization for the whole system as reported by the JVM
    Instrument: gauge
    Unit: 1
    """

    PROCESS_RUNTIME_JVM_SYSTEM_CPU_LOAD_1M = (
        "process.runtime.jvm.system.cpu.load_1m"
    )
    """
    Average CPU load of the whole system for the last minute as reported by the JVM
    Instrument: gauge
    Unit: 1
    """

    PROCESS_RUNTIME_JVM_BUFFER_USAGE = "process.runtime.jvm.buffer.usage"
    """
    Measure of memory used by buffers
    Instrument: updowncounter
    Unit: By
    """

    PROCESS_RUNTIME_JVM_BUFFER_LIMIT = "process.runtime.jvm.buffer.limit"
    """
    Measure of total memory capacity of buffers
    Instrument: updowncounter
    Unit: By
    """

    PROCESS_RUNTIME_JVM_BUFFER_COUNT = "process.runtime.jvm.buffer.count"
    """
    Number of buffers in the pool
    Instrument: updowncounter
    Unit: {buffer}
    """

    PROCESS_RUNTIME_JVM_MEMORY_USAGE = "process.runtime.jvm.memory.usage"
    """
    Measure of memory used
    Instrument: updowncounter
    Unit: By
    """

    PROCESS_RUNTIME_JVM_MEMORY_COMMITTED = (
        "process.runtime.jvm.memory.committed"
    )
    """
    Measure of memory committed
    Instrument: updowncounter
    Unit: By
    """

    PROCESS_RUNTIME_JVM_MEMORY_LIMIT = "process.runtime.jvm.memory.limit"
    """
    Measure of max obtainable memory
    Instrument: updowncounter
    Unit: By
    """

    PROCESS_RUNTIME_JVM_MEMORY_USAGE_AFTER_LAST_GC = (
        "process.runtime.jvm.memory.usage_after_last_gc"
    )
    """
    Measure of memory used, as measured after the most recent garbage collection event on this pool
    Instrument: updowncounter
    Unit: By
    """

    PROCESS_RUNTIME_JVM_GC_DURATION = "process.runtime.jvm.gc.duration"
    """
    Duration of JVM garbage collection actions
    Instrument: histogram
    Unit: s
    """

    PROCESS_RUNTIME_JVM_THREADS_COUNT = "process.runtime.jvm.threads.count"
    """
    Number of executing platform threads
    Instrument: updowncounter
    Unit: {thread}
    """

    PROCESS_RUNTIME_JVM_CLASSES_LOADED = "process.runtime.jvm.classes.loaded"
    """
    Number of classes loaded since JVM start
    Instrument: counter
    Unit: {class}
    """

    PROCESS_RUNTIME_JVM_CLASSES_UNLOADED = (
        "process.runtime.jvm.classes.unloaded"
    )
    """
    Number of classes unloaded since JVM start
    Instrument: counter
    Unit: {class}
    """

    PROCESS_RUNTIME_JVM_CLASSES_CURRENT_LOADED = (
        "process.runtime.jvm.classes.current_loaded"
    )
    """
    Number of classes currently loaded
    Instrument: updowncounter
    Unit: {class}
    """

    PROCESS_RUNTIME_JVM_CPU_TIME = "process.runtime.jvm.cpu.time"
    """
    CPU time used by the process as reported by the JVM
    Instrument: counter
    Unit: s
    """

    PROCESS_RUNTIME_JVM_CPU_RECENT_UTILIZATION = (
        "process.runtime.jvm.cpu.recent_utilization"
    )
    """
    Recent CPU utilization for the process as reported by the JVM
    Instrument: gauge
    Unit: 1
    """

    # Manually defined metrics

    DB_CLIENT_CONNECTIONS_USAGE = "db.client.connections.usage"
    """
    The number of connections that are currently in state described by the `state` attribute
    Instrument: UpDownCounter
    Unit: {connection}
    """



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""OpenTelemetry SDK File Configuration.

This module provides support for configuring the OpenTelemetry SDK
using declarative configuration files (YAML or JSON).

Example:
    >>> from opentelemetry.sdk._configuration.file import load_config_file
    >>> config = load_config_file("otel-config.yaml")
    >>> print(config.file_format)
    '1.0'
"""

from opentelemetry.sdk._configuration._exceptions import ConfigurationError
from opentelemetry.sdk._configuration._logger_provider import (
    configure_logger_provider,
    create_logger_provider,
)
from opentelemetry.sdk._configuration._meter_provider import (
    configure_meter_provider,
    create_meter_provider,
)
from opentelemetry.sdk._configuration._propagator import (
    configure_propagator,
    create_propagator,
)
from opentelemetry.sdk._configuration._resource import create_resource
from opentelemetry.sdk._configuration._tracer_provider import (
    configure_tracer_provider,
    create_tracer_provider,
)
from opentelemetry.sdk._configuration.file._env_substitution import (
    EnvSubstitutionError,
    substitute_env_vars,
)
from opentelemetry.sdk._configuration.file._loader import load_config_file

__all__ = [
    "load_config_file",
    "substitute_env_vars",
    "ConfigurationError",
    "EnvSubstitutionError",
    "create_resource",
    "create_propagator",
    "configure_propagator",
    "create_logger_provider",
    "configure_logger_provider",
    "create_tracer_provider",
    "configure_tracer_provider",
    "create_meter_provider",
    "configure_meter_provider",
]



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

SESSION_ID: Final = "session.id"
"""
A unique id to identify a session.
"""

SESSION_PREVIOUS_ID: Final = "session.previous_id"
"""
The previous `session.id` for this user, when known.
"""



"""
Execute a command and read the output as YAML. The YAML data is then directly overlaid onto the minion's Pillar data
"""

import logging

import salt.utils.yaml

# Don't "fix" the above docstring to put it on two lines, as the sphinx
# autosummary pulls only the first line for its description.


# Set up logging
log = logging.getLogger(__name__)


def ext_pillar(
    minion_id, pillar, command  # pylint: disable=W0613  # pylint: disable=W0613
):
    """
    Execute a command and read the output as YAML
    """
    try:
        command = command.replace("%s", minion_id)
        output = __salt__["cmd.run_stdout"](command, python_shell=True)
        return salt.utils.yaml.safe_load(output)
    except Exception:  # pylint: disable=broad-except
        log.critical(
            "YAML data from '%s' failed to parse. Command output:\n%s", command, output
        )
        return {}



r"""
A salt util for modifying the audit policies on the machine. This util is used
by the ``win_auditpol`` and ``win_lgpo`` modules.

Though this utility does not set group policy for auditing, it displays how all
auditing configuration is applied on the machine, either set directly or via
local or domain group policy.

.. versionadded:: 2018.3.4
.. versionadded:: 2019.2.1

Audit policy is read and written using the Windows ``advapi32`` APIs
(``AuditQuerySystemPolicy`` / ``AuditSetSystemPolicy``) with a static
subcategory GUID to English name map, so behavior is consistent regardless of
the host OS display language.

This util allows you to view and modify the audit settings as they are applied
on the machine. The audit settings are broken down into nine categories:

- Account Logon
- Account Management
- Detailed Tracking
- DS Access
- Logon/Logoff
- Object Access
- Policy Change
- Privilege Use
- System

The ``get_settings`` function will return the subcategories for all nine of
the above categories in one dictionary along with their auditing status.

Subcategory names are **canonical English** strings (they match the names Salt
and LGPO use, not the host OS localized ``auditpol`` display names).

To modify a setting you only need to specify the subcategory name and the value
you wish to set. Valid settings are:

- No Auditing
- Success
- Failure
- Success and Failure

The module constant ``settings`` maps those English labels to the integer
**auditing bitmask** passed to ``AuditSetSystemPolicy`` (``0``–``3``, aligned
with LGPO ``audit.csv`` ``Setting Value``). Execution modules should keep
using the string labels; callers should not rely on ``settings`` values being
``auditpol.exe`` switch strings.

LGPO loads defaults via :func:`get_advaudit_policy_rows`; :func:`get_auditpol_dump`
serializes the same data as UTF-8 CSV lines for backward compatibility.

Usage:

.. code-block:: python

    import salt.utils.win_lgpo_auditpol

    # Get current state of all audit settings
    salt.utils.win_lgpo_auditpol.get_settings()

    # Get the current state of all audit settings in the "Account Logon"
    # category
    salt.utils.win_lgpo_auditpol.get_settings(category="Account Logon")

    # Get current state of the "Credential Validation" setting
    salt.utils.win_lgpo_auditpol.get_setting(name='Credential Validation')

    # Set the state of the "Credential Validation" setting to Success and
    # Failure
    salt.utils.win_lgpo_auditpol.set_setting(name='Credential Validation',
                                             value='Success and Failure')

    # Set the state of the "Credential Validation" setting to No Auditing
    salt.utils.win_lgpo_auditpol.set_setting(name='Credential Validation',
                                             value='No Auditing')
"""

from __future__ import annotations

import contextlib
import csv
import ctypes
import io
import logging
import struct
import uuid
from ctypes import wintypes as w
from types import SimpleNamespace

import salt.utils.platform
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)
__virtualname__ = "auditpol"
__context__ = {}

categories = [
    "Account Logon",
    "Account Management",
    "Detailed Tracking",
    "DS Access",
    "Logon/Logoff",
    "Object Access",
    "Policy Change",
    "Privilege Use",
    "System",
]

# English value label -> bitmask for AuditSetSystemPolicy (matches LGPO 0–3)
settings = {
    "No Auditing": 0,
    "Success": 1,
    "Failure": 2,
    "Success and Failure": 3,
}

_AUDIT_NONE = 0
_AUDIT_SUCCESS = 1
_AUDIT_FAILURE = 2

_TOKEN_QUERY = 0x0008
_TOKEN_ADJUST_PRIVILEGES = 0x0020
_SE_PRIVILEGE_ENABLED = 0x00000002

_FIELDNAMES = [
    "Machine Name",
    "Policy Target",
    "Subcategory",
    "Subcategory GUID",
    "Inclusion Setting",
    "Exclusion Setting",
    "Setting Value",
]


class _GUID(ctypes.Structure):
    """Win32 ``GUID`` layout (``Data1``..``Data4``) for ``Audit*`` APIs."""

    _fields_ = [
        ("Data1", w.DWORD),
        ("Data2", w.WORD),
        ("Data3", w.WORD),
        ("Data4", w.BYTE * 8),
    ]


class _AUDIT_POLICY_INFORMATION(ctypes.Structure):
    """``AUDIT_POLICY_INFORMATION`` from ntsecapi / advapi32 (per subcategory)."""

    _fields_ = [
        ("AuditSubcategoryGuid", _GUID),
        ("AuditingInformation", w.ULONG),
        ("AuditCategoryGuid", _GUID),
    ]


class _LUID(ctypes.Structure):
    """Locally unique identifier (used with ``LookupPrivilegeValueW``)."""

    _fields_ = [("LowPart", w.DWORD), ("HighPart", w.LONG)]


class _LUID_AND_ATTRIBUTES(ctypes.Structure):
    """One privilege entry for ``TOKEN_PRIVILEGES``."""

    _fields_ = [("Luid", _LUID), ("Attributes", w.DWORD)]


class _TOKEN_PRIVILEGES(ctypes.Structure):
    """``TOKEN_PRIVILEGES`` with a single ``LUID_AND_ATTRIBUTES`` (fixed size)."""

    _fields_ = [("PrivilegeCount", w.DWORD), ("Privileges", _LUID_AND_ATTRIBUTES * 1)]


def _load_advapi32():
    """
    Bind advapi32/kernel32 entry points used for audit policy and token
    privileges. ``use_last_error=True`` so ``ctypes.get_last_error()`` matches
    Win32 failures after each call.
    """
    advapi32_dll = ctypes.WinDLL("advapi32", use_last_error=True)
    kernel32_dll = ctypes.WinDLL("kernel32", use_last_error=True)

    # AuditQuerySystemPolicy — ntsecapi.h; ppAuditPolicy is heap-allocated; free
    # with AuditFree. Local names match Win32 exports for easier MSDN lookup.
    AuditQuerySystemPolicy = advapi32_dll.AuditQuerySystemPolicy
    AuditQuerySystemPolicy.argtypes = [
        ctypes.POINTER(_GUID),
        w.ULONG,
        ctypes.POINTER(ctypes.POINTER(_AUDIT_POLICY_INFORMATION)),
    ]
    AuditQuerySystemPolicy.restype = w.BOOL

    # AuditSetSystemPolicy — category GUID member ignored per MSDN.
    AuditSetSystemPolicy = advapi32_dll.AuditSetSystemPolicy
    AuditSetSystemPolicy.argtypes = [
        ctypes.POINTER(_AUDIT_POLICY_INFORMATION),
        w.ULONG,
    ]
    AuditSetSystemPolicy.restype = w.BOOL

    AuditFree = advapi32_dll.AuditFree
    AuditFree.argtypes = [w.LPVOID]
    AuditFree.restype = None

    OpenProcessToken = advapi32_dll.OpenProcessToken
    OpenProcessToken.argtypes = [w.HANDLE, w.DWORD, ctypes.POINTER(w.HANDLE)]
    OpenProcessToken.restype = w.BOOL

    LookupPrivilegeValueW = advapi32_dll.LookupPrivilegeValueW
    LookupPrivilegeValueW.argtypes = [
        w.LPCWSTR,
        w.LPCWSTR,
        ctypes.POINTER(_LUID),
    ]
    LookupPrivilegeValueW.restype = w.BOOL

    AdjustTokenPrivileges = advapi32_dll.AdjustTokenPrivileges
    AdjustTokenPrivileges.argtypes = [
        w.HANDLE,
        w.BOOL,
        ctypes.POINTER(_TOKEN_PRIVILEGES),
        w.DWORD,
        ctypes.c_void_p,
        ctypes.c_void_p,
    ]
    AdjustTokenPrivileges.restype = w.BOOL

    GetCurrentProcess = kernel32_dll.GetCurrentProcess
    GetCurrentProcess.argtypes = []
    GetCurrentProcess.restype = w.HANDLE

    CloseHandle = kernel32_dll.CloseHandle
    CloseHandle.argtypes = [w.HANDLE]
    CloseHandle.restype = w.BOOL

    return SimpleNamespace(
        AuditQuerySystemPolicy=AuditQuerySystemPolicy,
        AuditSetSystemPolicy=AuditSetSystemPolicy,
        AuditFree=AuditFree,
        OpenProcessToken=OpenProcessToken,
        LookupPrivilegeValueW=LookupPrivilegeValueW,
        AdjustTokenPrivileges=AdjustTokenPrivileges,
        GetCurrentProcess=GetCurrentProcess,
        CloseHandle=CloseHandle,
    )


_API = None


def _api():
    """
    Lazy singleton of :func:`_load_advapi32` bindings (``_API`` cache).

    Attributes use **PascalCase** names matching the Win32 exports (e.g.
    ``AuditQuerySystemPolicy``) for parity with MSDN and headers.
    """
    global _API
    if _API is None:
        _API = _load_advapi32()
    return _API


def _uuid_to_guid(subcategory_uuid: uuid.UUID) -> _GUID:
    """
    Pack a Python :class:`uuid.UUID` into the Win32 ``GUID`` memory layout.

    ``UUID.bytes_le`` matches how ``GUID`` fields are ordered in RAM for the
    ``Data1``/``Data2``/``Data3``/``Data4`` split used by ``AuditQuerySystemPolicy``.
    """
    uuid_bytes_le = subcategory_uuid.bytes_le
    guid_struct = _GUID()
    guid_struct.Data1, guid_struct.Data2, guid_struct.Data3 = struct.unpack(
        "<IHH", uuid_bytes_le[:8]
    )
    ctypes.memmove(guid_struct.Data4, uuid_bytes_le[8:16], 8)
    return guid_struct


def _mask_to_labels(audit_mask: int) -> tuple[str, str]:
    """Map ``AuditingInformation`` bitmask to (Inclusion Setting, Setting Value str)."""
    success_on = bool(audit_mask & _AUDIT_SUCCESS)
    failure_on = bool(audit_mask & _AUDIT_FAILURE)
    if success_on and failure_on:
        return "Success and Failure", "3"
    if success_on:
        return "Success", "1"
    if failure_on:
        return "Failure", "2"
    return "No Auditing", "0"


@contextlib.contextmanager
def _enable_se_security_privilege():
    """
    Enable ``SeSecurityPrivilege`` on this process for the duration of the
    ``with`` block (required for ``AuditSetSystemPolicy``).
    """
    win32 = _api()
    process_token_handle = w.HANDLE()
    if not win32.OpenProcessToken(
        win32.GetCurrentProcess(),
        _TOKEN_ADJUST_PRIVILEGES | _TOKEN_QUERY,
        ctypes.byref(process_token_handle),
    ):
        raise CommandExecutionError(
            "OpenProcessToken failed", info={"errno": ctypes.get_last_error()}
        )
    try:
        security_privilege_luid = _LUID()
        if not win32.LookupPrivilegeValueW(
            None, "SeSecurityPrivilege", ctypes.byref(security_privilege_luid)
        ):
            raise CommandExecutionError(
                "LookupPrivilegeValueW(SeSecurityPrivilege) failed",
                info={"errno": ctypes.get_last_error()},
            )
        token_privileges = _TOKEN_PRIVILEGES()
        token_privileges.PrivilegeCount = 1
        token_privileges.Privileges[0].Luid = security_privilege_luid
        token_privileges.Privileges[0].Attributes = _SE_PRIVILEGE_ENABLED
        if not win32.AdjustTokenPrivileges(
            process_token_handle,
            False,
            ctypes.byref(token_privileges),
            ctypes.sizeof(token_privileges),
            None,
            None,
        ):
            raise CommandExecutionError(
                "AdjustTokenPrivileges failed",
                info={"errno": ctypes.get_last_error()},
            )
        yield
    finally:
        win32.CloseHandle(process_token_handle)


def _query_system_policies():
    """
    Call ``AuditQuerySystemPolicy`` for every subcategory in
    :data:`_AUDIT_SUBCATEGORY_METADATA`.

    Returns:
        list: One tuple per row:
        ``(category_name, subcategory_name, subcategory_uuid, audit_mask)`` —
        ``audit_mask`` is the raw ``AuditingInformation`` ULONG (success/failure
        bits per MSDN).
    """
    metadata_rows = _AUDIT_SUBCATEGORY_METADATA
    subcategory_count = len(metadata_rows)

    # Contiguous array of GUIDs — required; passing a single GUID pointer is a
    # common marshaling mistake and yields ERROR_INVALID_PARAMETER.
    guid_array_type = _GUID * subcategory_count
    subcategory_guid_array = guid_array_type()
    subcategory_uuids_ordered = []
    for index, (_category, _subcategory, guid_string) in enumerate(metadata_rows):
        parsed_uuid = uuid.UUID(guid_string)
        subcategory_uuids_ordered.append(parsed_uuid)
        subcategory_guid_array[index] = _uuid_to_guid(parsed_uuid)

    # Output: pointer to heap array of AUDIT_POLICY_INFORMATION (same length).
    allocated_policy_array_ptr = ctypes.POINTER(_AUDIT_POLICY_INFORMATION)()
    win32 = _api()
    if not win32.AuditQuerySystemPolicy(
        subcategory_guid_array,
        subcategory_count,
        ctypes.byref(allocated_policy_array_ptr),
    ):
        err = ctypes.get_last_error()
        raise CommandExecutionError(
            "AuditQuerySystemPolicy failed",
            info={"errno": err},
        )
    if not allocated_policy_array_ptr:
        return [
            (
                metadata_rows[i][0],
                metadata_rows[i][1],
                subcategory_uuids_ordered[i],
                0,
            )
            for i in range(subcategory_count)
        ]
    try:
        results = []
        for index in range(subcategory_count):
            policy_info_struct = allocated_policy_array_ptr[index]
            results.append(
                (
                    metadata_rows[index][0],
                    metadata_rows[index][1],
                    subcategory_uuids_ordered[index],
                    int(policy_info_struct.AuditingInformation),
                )
            )
        return results
    finally:
        win32.AuditFree(allocated_policy_array_ptr)


# (category, subcategory English name as in auditpol backup / set, GUID string)
# GUIDs align with Microsoft advanced audit subcategories (see e.g. PowerShell
# AuditPolicyDsc AuditPolicyResourceHelper).
_AUDIT_SUBCATEGORY_METADATA = [
    ("System", "Security State Change", "{0CCE9210-69AE-11D9-BED3-505054503030}"),
    ("System", "Security System Extension", "{0CCE9211-69AE-11D9-BED3-505054503030}"),
    ("System", "System Integrity", "{0CCE9212-69AE-11D9-BED3-505054503030}"),
    ("System", "IPsec Driver", "{0CCE9213-69AE-11D9-BED3-505054503030}"),
    ("System", "Other System Events", "{0CCE9214-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "Logon", "{0CCE9215-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "Logoff", "{0CCE9216-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "Account Lockout", "{0CCE9217-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "IPsec Main Mode", "{0CCE9218-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "IPsec Quick Mode", "{0CCE9219-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "IPsec Extended Mode", "{0CCE921A-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "Special Logon", "{0CCE921B-69AE-11D9-BED3-505054503030}"),
    (
        "Logon/Logoff",
        "Other Logon/Logoff Events",
        "{0CCE921C-69AE-11D9-BED3-505054503030}",
    ),
    ("Logon/Logoff", "Network Policy Server", "{0CCE9243-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "User / Device Claims", "{0CCE9247-69AE-11D9-BED3-505054503030}"),
    ("Logon/Logoff", "Group Membership", "{0CCE9249-69AE-11D9-BED3-505054503030}"),
    ("Object Access", "File System", "{0CCE921D-69AE-11D9-BED3-505054503030}"),
    ("Object Access", "Registry", "{0CCE921E-69AE-11D9-BED3-505054503030}"),
    ("Object Access", "Kernel Object", "{0CCE921F-69AE-11D9-BED3-505054503030}"),
    ("Object Access", "SAM", "{0CCE9220-69AE-11D9-BED3-505054503030}"),
    (
        "Object Access",
        "Certification Services",
        "{0CCE9221-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Object Access",
        "Application Generated",
        "{0CCE9222-69AE-11D9-BED3-505054503030}",
    ),
    ("Object Access", "Handle Manipulation", "{0CCE9223-69AE-11D9-BED3-505054503030}"),
    ("Object Access", "File Share", "{0CCE9224-69AE-11D9-BED3-505054503030}"),
    (
        "Object Access",
        "Filtering Platform Packet Drop",
        "{0CCE9225-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Object Access",
        "Filtering Platform Connection",
        "{0CCE9226-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Object Access",
        "Other Object Access Events",
        "{0CCE9227-69AE-11D9-BED3-505054503030}",
    ),
    ("Object Access", "Detailed File Share", "{0CCE9244-69AE-11D9-BED3-505054503030}"),
    ("Object Access", "Removable Storage", "{0CCE9245-69AE-11D9-BED3-505054503030}"),
    (
        "Object Access",
        "Central Policy Staging",
        "{0CCE9246-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Privilege Use",
        "Sensitive Privilege Use",
        "{0CCE9228-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Privilege Use",
        "Non Sensitive Privilege Use",
        "{0CCE9229-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Privilege Use",
        "Other Privilege Use Events",
        "{0CCE922A-69AE-11D9-BED3-505054503030}",
    ),
    ("Detailed Tracking", "Process Creation", "{0CCE922B-69AE-11D9-BED3-505054503030}"),
    (
        "Detailed Tracking",
        "Process Termination",
        "{0CCE922C-69AE-11D9-BED3-505054503030}",
    ),
    ("Detailed Tracking", "DPAPI Activity", "{0CCE922D-69AE-11D9-BED3-505054503030}"),
    ("Detailed Tracking", "RPC Events", "{0CCE922E-69AE-11D9-BED3-505054503030}"),
    (
        "Detailed Tracking",
        "Plug and Play Events",
        "{0CCE9248-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Detailed Tracking",
        "Token Right Adjusted Events",
        "{0CCE924A-69AE-11D9-BED3-505054503030}",
    ),
    ("Policy Change", "Audit Policy Change", "{0CCE922F-69AE-11D9-BED3-505054503030}"),
    (
        "Policy Change",
        "Authentication Policy Change",
        "{0CCE9230-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Policy Change",
        "Authorization Policy Change",
        "{0CCE9231-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Policy Change",
        "MPSSVC Rule-Level Policy Change",
        "{0CCE9232-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Policy Change",
        "Filtering Platform Policy Change",
        "{0CCE9233-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Policy Change",
        "Other Policy Change Events",
        "{0CCE9234-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Management",
        "User Account Management",
        "{0CCE9235-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Management",
        "Computer Account Management",
        "{0CCE9236-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Management",
        "Security Group Management",
        "{0CCE9237-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Management",
        "Distribution Group Management",
        "{0CCE9238-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Management",
        "Application Group Management",
        "{0CCE9239-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Management",
        "Other Account Management Events",
        "{0CCE923A-69AE-11D9-BED3-505054503030}",
    ),
    ("DS Access", "Directory Service Access", "{0CCE923B-69AE-11D9-BED3-505054503030}"),
    (
        "DS Access",
        "Directory Service Changes",
        "{0CCE923C-69AE-11D9-BED3-505054503030}",
    ),
    (
        "DS Access",
        "Directory Service Replication",
        "{0CCE923D-69AE-11D9-BED3-505054503030}",
    ),
    (
        "DS Access",
        "Detailed Directory Service Replication",
        "{0CCE923E-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Logon",
        "Credential Validation",
        "{0CCE923F-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Logon",
        "Kerberos Service Ticket Operations",
        "{0CCE9240-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Logon",
        "Other Account Logon Events",
        "{0CCE9241-69AE-11D9-BED3-505054503030}",
    ),
    (
        "Account Logon",
        "Kerberos Authentication Service",
        "{0CCE9242-69AE-11D9-BED3-505054503030}",
    ),
]


def __virtual__():
    """
    Only load if on a Windows system
    """
    if not salt.utils.platform.is_windows():
        return False, "This utility only available on Windows"

    return __virtualname__


def get_advaudit_policy_rows():
    """
    Return one row per advanced audit subcategory for LGPO and related code.

    Rows are built from ``AuditQuerySystemPolicy`` and the in-module GUID/name
    table, so they do not depend on ``auditpol.exe`` or the host display
    language.

    Returns:
        list: A list of dictionaries, each with keys ``Machine Name``,
            ``Policy Target``, ``Subcategory``, ``Subcategory GUID``,
            ``Inclusion Setting``, ``Exclusion Setting``, and ``Setting Value``.
            ``Machine Name`` is always ``""``; ``Policy Target`` is ``System``;
            ``Subcategory`` is the English name used with :func:`set_setting`;
            ``Subcategory GUID`` is braced uppercase; ``Setting Value`` is
            ``"0"``–``"3"`` matching LGPO's ``audit.csv`` convention.

    Raises:
        CommandExecutionError: If the Windows API call fails.
    """
    rows_out = []
    for (_metadata_category, subcategory, guid_string), query_row in zip(
        _AUDIT_SUBCATEGORY_METADATA, _query_system_policies()
    ):
        _, _, _, audit_mask = query_row
        inclusion, setting_val = _mask_to_labels(audit_mask)
        guid_braced = guid_string.upper()
        rows_out.append(
            {
                "Machine Name": "",
                "Policy Target": "System",
                "Subcategory": subcategory,
                "Subcategory GUID": guid_braced,
                "Inclusion Setting": inclusion,
                "Exclusion Setting": "",
                "Setting Value": setting_val,
            }
        )
    return rows_out


def get_settings(category="All"):
    """
    Get the current configuration for all audit settings in the given category.

    Reads effective policy via ``AuditQuerySystemPolicy`` (not ``auditpol
    /get``), so results use English subcategory keys and English value labels
    regardless of OS language.

    Args:
        category (str):
            One of the nine categories, or ``All`` / ``*`` for every
            subcategory. Names are matched case-insensitively against
            :data:`categories`.

    Returns:
        dict: Maps each **English** subcategory name to one of ``No Auditing``,
            ``Success``, ``Failure``, or ``Success and Failure``.

    Raises:
        KeyError: If ``category`` is not recognized.
        CommandExecutionError: If the Windows API call fails.
    """
    if category.lower() in ["all", "*"]:
        want = None
    else:
        want = None
        for c in categories:
            if c.lower() == category.lower():
                want = c
                break
        if want is None:
            raise KeyError(f'Invalid category: "{category}"')

    ret = {}
    for (category_name, subcategory_name, _subcategory_uuid), query_row in zip(
        _AUDIT_SUBCATEGORY_METADATA, _query_system_policies()
    ):
        _, _, _, audit_mask = query_row
        if want is not None and category_name != want:
            continue
        label, _setting_value_str = _mask_to_labels(audit_mask)
        ret[subcategory_name] = label
    return ret


def get_setting(name):
    """
    Get the current configuration for a single subcategory.

    Args:
        name (str): English subcategory name (matched case-insensitively).

    Returns:
        str: One of ``No Auditing``, ``Success``, ``Failure``, or
            ``Success and Failure``.

    Raises:
        KeyError: If ``name`` is not a known subcategory.
        CommandExecutionError: If querying the current policy fails.
    """
    current_settings = get_settings(category="All")
    for setting in current_settings:
        if name.lower() == setting.lower():
            return current_settings[setting]
    raise KeyError(f"Invalid name: {name}")


def _get_valid_names():
    """
    Return lowercase English subcategory names valid for :func:`set_setting`.

    Cached on ``__context__['auditpol.valid_names']`` until cleared after a
    successful :func:`set_setting`.
    """
    if "auditpol.valid_names" not in __context__:
        settings_map = get_settings(category="All")
        __context__["auditpol.valid_names"] = [k.lower() for k in settings_map]
    return __context__["auditpol.valid_names"]


def set_setting(name, value):
    """
    Set the auditing bitmask for one subcategory.

    Enables ``SeSecurityPrivilege`` on the current process token for the
    duration of the ``AuditSetSystemPolicy`` call.

    Args:
        name (str):
            English subcategory name (same strings as returned by
            :func:`get_settings`).
        value (str):
            One of ``No Auditing``, ``Success``, ``Failure``, or
            ``Success and Failure`` (matched case-insensitively).

    Returns:
        bool: ``True`` on success.

    Raises:
        KeyError: On invalid ``name`` or ``value``.
        CommandExecutionError: If privilege adjustment or
            ``AuditSetSystemPolicy`` fails.
    """
    if name.lower() not in _get_valid_names():
        raise KeyError(f"Invalid name: {name}")
    audit_bitmask = None
    for setting in settings:
        if value.lower() == setting.lower():
            audit_bitmask = int(settings[setting])
            break
    if audit_bitmask is None:
        raise KeyError(f"Invalid setting value: {value}")

    resolved_subcategory_name = None
    resolved_subcategory_uuid = None
    for _meta_cat, meta_subcategory, guid_string in _AUDIT_SUBCATEGORY_METADATA:
        if name.lower() == meta_subcategory.lower():
            resolved_subcategory_name = meta_subcategory
            resolved_subcategory_uuid = uuid.UUID(guid_string)
            break
    if resolved_subcategory_uuid is None:
        raise KeyError(f"Invalid name: {name}")

    audit_policy_info = _AUDIT_POLICY_INFORMATION()
    audit_policy_info.AuditSubcategoryGuid = _uuid_to_guid(resolved_subcategory_uuid)
    audit_policy_info.AuditingInformation = audit_bitmask
    audit_policy_info.AuditCategoryGuid = _GUID()

    win32 = _api()
    with _enable_se_security_privilege():
        policy_count = 1
        if not win32.AuditSetSystemPolicy(
            ctypes.byref(audit_policy_info), policy_count
        ):
            err = ctypes.get_last_error()
            raise CommandExecutionError(
                "AuditSetSystemPolicy failed",
                info={
                    "errno": err,
                    "name": resolved_subcategory_name,
                    "value": value,
                },
            )

    __context__.pop("auditpol.valid_names", None)
    return True


def get_auditpol_dump():
    """
    Return advanced audit policy rows as CSV **text lines** (UTF-8).

    This does not run ``auditpol /backup`` or read a temp file. It formats the
    same data as :func:`get_advaudit_policy_rows` using :mod:`csv`, so each line
    is a normal Unicode string (newlines ``\\n``).

    Returns:
        list: Lines including a header row, suitable for passing to
            :class:`csv.DictReader` if needed.

    Raises:
        CommandExecutionError: If building rows fails (e.g. API error from
            :func:`get_advaudit_policy_rows`).
    """
    buf = io.StringIO(newline="")
    writer = csv.DictWriter(buf, fieldnames=_FIELDNAMES, lineterminator="\n")
    writer.writeheader()
    for row in get_advaudit_policy_rows():
        writer.writerow(row)
    buf.seek(0)
    return buf.readlines()



# This file is a Salt-maintained security patch of pip's vendored urllib3.
#
# The underlying code is urllib3 1.26.20 (the version vendored by pip 25.2)
# with the following CVE fixes backported from upstream urllib3 2.6.3:
#
#   CVE-2025-66418 (GHSA-gm62-xv2j-4w53): Unbounded Content-Encoding
#     decompression chain -- MultiDecoder now enforces a 5-link limit.
#     Upstream fix: urllib3 2.6.0 (commit 24d7b67).
#
#   CVE-2026-21441 (GHSA-38jv-5279-wg99): drain_conn unnecessarily
#     decompressed the full body of HTTP redirect responses, creating a
#     decompression-bomb vector. Fixed by adding _has_decoded_content
#     tracking and only decoding in drain_conn when decoding was already
#     in progress.
#     Upstream fix: urllib3 2.6.3 (commit 8864ac4).
#
#   CVE-2025-66471 (GHSA-2xpw-w6gg-jr37): Decompression bomb in the
#     streaming API via max_length parameter. NOT backported -- requires a
#     full 2.x streaming infrastructure refactor. Ubuntu did not backport
#     this to 1.26.x either. pip maintainers confirmed pip is not
#     affected because all pip network calls use decode_content=False.
#
# The version string "2.6.3" reflects the highest upstream release from
# which fixes have been backported. The underlying API remains urllib3
# 1.26.x -- this is NOT a port to urllib3 2.x.
__version__ = "2.6.3"
